"""Calculate metricity of a bar"""

import functools
import math
import operator
import typing

from mutwo.core import converters
from mutwo.core.utilities import prime_factors

__all__ = ("RhythmicalStrataToIndispensabilityConverter",)


class RhythmicalStrataToIndispensabilityConverter(converters.abc.Converter):
    """Builds metrical indispensability for a rhythmical strata.

    This technique has been described by Clarence Barlow in `On the Quantification
    of Harmony and Metre` (1992). The technique aims to model the weight
    of single beats in a particular metre. It allocates each beat of a metre
    to a specific value that describes the `indispensability` of a beat: the higher the
    assigned value, the more accented the beat.
    """

    # ###################################################################### #
    #                          static methods                                #
    # ###################################################################### #

    @staticmethod
    def _indispensability_of_nth_beat_for_simple_meter(
        beat_index: int, prime_number: int
    ) -> int:
        """Calculate indispensability for nth beat of a metre that is composed of a singular prime number."""

        if prime_number == 2:
            return prime_number - beat_index
        elif beat_index == prime_number - 1:
            return int(prime_number / 4)
        else:
            factorised = tuple(
                sorted(prime_factors.factorise(prime_number - 1), reverse=True)
            )
            q = RhythmicalStrataToIndispensabilityConverter._indispensability_of_nth_beat(
                beat_index - int(beat_index / prime_number), factorised
            )
            return int(q + (2 * math.sqrt((q + 1) / prime_number)))

    @staticmethod
    def _indispensability_of_nth_beat(
        beat_index: int, rhythmical_strata: tuple[int, ...]
    ) -> int:
        """Calculate indispensability for the nth beat of a metre.

        The metre is defined with the primes argument.
        """

        z = len(rhythmical_strata)
        rhythmical_strata = (1,) + rhythmical_strata + (1,)
        r_sum = []
        for r in range(0, z):
            up = (beat_index - 2) % functools.reduce(
                operator.mul, tuple(rhythmical_strata[j] for j in range(1, z + 1))
            )
            down = functools.reduce(
                operator.mul, tuple(rhythmical_strata[z + 1 - k] for k in range(r + 1))
            )
            local_result = 1 + (int(1 + (up / down)) % rhythmical_strata[z - r])
            base_indispensability = RhythmicalStrataToIndispensabilityConverter._indispensability_of_nth_beat_for_simple_meter(
                local_result, rhythmical_strata[z - r]
            )
            product = functools.reduce(
                operator.mul, tuple(rhythmical_strata[i] for i in range(0, z - r))
            )
            r_sum.append(product * base_indispensability)
        return sum(r_sum)

    # ###################################################################### #
    #               public methods for interaction with the user             #
    # ###################################################################### #

    def convert(
        self, rhythmical_strata_to_convert: typing.Sequence[int]
    ) -> tuple[int, ...]:
        """Convert indispensability for each beat of a particular metre.

        :param rhythmical_strata_to_convert: The rhythmical strata defines
            the metre for which the indispensability shall be calculated. The
            rhythmical strata is a list of prime numbers which product is the
            amount of available beats within the particular metre. Earlier
            prime numbers in the rhythmical strata are considered to be more
            important than later prime numbers.
        :return: A tuple of a integer for each beat of the respective metre where
            each integer describes how accented the particular beat is (the higher the
            number, the more important the beat).

        **Example:**

        >>> from mutwo.converters import symmetrical
        >>> metricity_converter = symmetrical.metricities.RhythmicalStrataToIndispensabilityConverter()
        >>> metricity_converter.convert((2, 3))  # time signature 3/4
        (5, 0, 3, 1, 4, 2)
        >>> metricity_converter.convert((3, 2))  # time signature 6/8
        (5, 0, 2, 4, 1, 3)
        """

        # inverse so that the first numbers are more important than the later numbers
        rhythmical_strata_to_convert = tuple(reversed(rhythmical_strata_to_convert))

        length = functools.reduce(operator.mul, rhythmical_strata_to_convert)
        return tuple(
            RhythmicalStrataToIndispensabilityConverter._indispensability_of_nth_beat(
                i + 1, rhythmical_strata_to_convert
            )
            for i in range(length)
        )



"""Submodule for the parameter pitch.

'Pitch' is defined as any object that knows a :attr:`frequency` attribute.
The two major modern tuning systems Just intonation and Equal-divided-octave
are supported by the :class:`JustIntonationPitch` and :class:`EqualDividedOctavePitch` classes.
For using Western nomenclature (e.g. c, d, e, f, ...) :mod:`mutwo` offers the
:class:`WesternPitch` class (which inherits from :class:`EqualDividedOctavePitch`).
For a straight frequency-based approach one may use :class:`DirectPitch`.

If desired the default concert pitch can be adjusted after importing :mod:`mutwo`:

    >>> from mutwo.parameters import pitches_constants
    >>> pitches_constants.DEFAULT_CONCERT_PITCH = 443

All pitch objects with a concert pitch attribute that become initialised after
overriding the default concert pitch value will by default use the new
overridden default concert pitch value.
"""

from __future__ import annotations

import bisect
import collections
import copy
import functools
import math
import numbers
import operator
import typing
import warnings

import primesieve  # type: ignore
from primesieve import numpy as primesieve_numpy

try:
    import quicktions as fractions  # type: ignore
except ImportError:
    import fractions  # type: ignore

from mutwo.core.utilities import constants
from mutwo.core.utilities import decorators
from mutwo.core.utilities import prime_factors
from mutwo.core.utilities import tools

from mutwo.core import parameters

__all__ = (
    "DirectPitch",
    "JustIntonationPitch",
    "EqualDividedOctavePitch",
    "WesternPitch",
    "MidiPitch",
    "CommonHarmonic",
)

ConcertPitch = typing.Union[constants.Real, parameters.abc.Pitch]
PitchClassOrPitchClassName = typing.Union[constants.Real, str]


class DirectPitchInterval(parameters.abc.PitchInterval):
    """Simple interval class which gets directly assigned by its cents value

    :param cents: Defines how big or small the interval is.
    :type cents: float
    """

    def __init__(self, cents: float):
        self.cents = cents

    @property
    def cents(self) -> float:
        return self._cents

    @cents.setter
    def cents(self, cents: float):
        self._cents = cents


class DirectPitch(parameters.abc.Pitch):
    """A simple pitch class that gets directly initialised by its frequency.

    :param frequency: The frequency of the ``DirectPitch`` object.

    May be used when a converter class needs a pitch object, but there is
    no need or desire for a complex abstraction of the respective pitch
    (that classes like ``JustIntonationPitch`` or ``WesternPitch`` offer).

    **Example:**

    >>> from mutwo.parameters import pitches
    >>> my_pitch = pitches.DirectPitch(440)
    """

    def __init__(self, frequency: constants.Real, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._frequency = float(frequency)

    @property
    def frequency(self) -> float:
        """The frequency of the pitch."""

        return self._frequency

    def __repr__(self) -> str:
        return "DirectPitch(frequency = {})".format(self.frequency)

    @decorators.add_copy_option
    def add(
        self, pitch_interval: parameters.abc.PitchInterval, mutate: bool = False
    ) -> DirectPitch:
        self._frequency = self.cents_to_ratio(pitch_interval.cents) * self.frequency
        return self

    @decorators.add_copy_option
    def subtract(
        self, pitch_interval: parameters.abc.PitchInterval, mutate: bool = False
    ) -> DirectPitch:
        self._frequency = self.frequency / self.cents_to_ratio(pitch_interval.cents)
        return self


class MidiPitch(parameters.abc.Pitch):
    """Pitch that is defined by its midi pitch number.

    :param midi_pitch_number: The midi pitch number of the pitch. Floating
        point numbers are possible for microtonal deviations from the
        chromatic scale.
    :type midi_pitch_number: float

    **Example:**

    >>> from mutwo.parameters import pitches
    >>> middle_c = pitches.MidiPitch(60)
    >>> middle_c_quarter_tone_high = pitches.MidiPitch(60.5)
    """

    def __init__(self, midi_pitch_number: float, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._midi_pitch_number = midi_pitch_number

    @property
    def frequency(self) -> float:
        difference_to_middle_a = self.midi_pitch_number - 69
        return float(440 * self.cents_to_ratio(difference_to_middle_a * 100))

    @property
    def midi_pitch_number(self) -> float:
        return self._midi_pitch_number

    @midi_pitch_number.setter
    def midi_pitch_number(self, new_midi_pitch_number: float):
        self._midi_pitch_number = new_midi_pitch_number

    @decorators.add_copy_option
    def add(
        self, pitch_interval: parameters.abc.PitchInterval, mutate: bool = False
    ) -> MidiPitch:
        self.midi_pitch_number = self.hertz_to_midi_pitch_number(
            self.cents_to_ratio(pitch_interval.cents) * self.frequency
        )
        return self

    @decorators.add_copy_option
    def subtract(
        self, pitch_interval: parameters.abc.PitchInterval, mutate: bool = False
    ) -> MidiPitch:
        self.midi_pitch_number = self.hertz_to_midi_pitch_number(
            self.frequency / self.cents_to_ratio(pitch_interval.cents)
        )
        return self


class JustIntonationPitch(parameters.abc.Pitch, parameters.abc.PitchInterval):
    """Pitch that is defined by a frequency ratio and a reference pitch.

    :param ratio_or_exponent_tuple: The frequency ratio of the ``JustIntonationPitch``.
        This can either be a string that indicates the frequency ratio (for
        instance: "1/1", "3/2", "9/2", etc.), or a ``fractions.Fraction``
        object that indicates the frequency ratio (for instance:
        ``fractions.Fraction(3, 2)``, ``fractions.Fraction(7, 4)``) or
        an Iterable that is filled with integer that represents the exponent_tuple
        of the respective prime numbers of the decomposed frequency ratio. The prime
        numbers are rising and start with 2. Therefore the tuple ``(2, 0, -1)``
        would return the frequency ratio ``4/5`` because
        ``(2 ** 2) * (3 ** 0) * (5 ** -1) = 4/5``.
    :param concert_pitch: The reference pitch of the tuning system (the pitch for a
        frequency ratio of 1/1). Can either be another ``Pitch`` object or any number
        to indicate a particular frequency in Hertz.

    The resulting frequency is calculated by multiplying the frequency ratio
    with the respective reference pitch.

    **Example:**

    >>> from mutwo.parameters import pitches
    >>> # 3 different variations of initialising the same pitch
    >>> pitches.JustIntonationPitch('3/2')
    >>> import fractions
    >>> pitches.JustIntonationPitch(fractions.Fraction(3, 2))
    >>> pitches.JustIntonationPitch((-1, 1))
    >>> # using a different concert pitch
    >>> pitches.JustIntonationPitch('7/5', concert_pitch=432)
    """

    def __init__(
        self,
        ratio_or_exponent_tuple: typing.Union[
            str, fractions.Fraction, typing.Iterable[int]
        ] = "1/1",
        concert_pitch: ConcertPitch = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if concert_pitch is None:
            concert_pitch = parameters.pitches_constants.DEFAULT_CONCERT_PITCH

        self.exponent_tuple = (
            self._translate_ratio_or_fractions_argument_to_exponent_tuple(
                ratio_or_exponent_tuple
            )
        )
        self.concert_pitch = concert_pitch  # type: ignore

    # ###################################################################### #
    #                      static private methods                            #
    # ###################################################################### #
    @staticmethod
    def _adjust_exponent_lengths(
        exponent_tuple0: tuple, exponent_tuple1: tuple
    ) -> tuple:
        r"""Adjust two exponent_tuple, e.g. make their length equal.

        The length of the longer JustIntonationPitch is the reference.

        Arguments:
            * exponent_tuple0: first exponent_tuple to adjust
            * exponent_tuple1: second exponent_tuple to adjust
        >>> v0 = (1, 0, -1)
        >>> v1 = (1,)
        >>> v0_adjusted, v1_adjusted = JustIntonationPitch._adjust_exponent_lengths(v0, v1)
        >>> v0_adjusted
        (1, 0, -1)
        >>> v1_adjusted
        (1, 0, 0)
        """

        length0 = len(exponent_tuple0)
        length1 = len(exponent_tuple1)
        if length0 > length1:
            return exponent_tuple0, exponent_tuple1 + (0,) * (length0 - length1)
        else:
            return exponent_tuple0 + (0,) * (length1 - length0), exponent_tuple1

    @staticmethod
    def _adjust_ratio(ratio: fractions.Fraction, border: int) -> fractions.Fraction:
        r"""Multiply or divide a fractions.Fraction - Object with the border,

        until it is equal or bigger than 1 and smaller than border.

        Arguments:
            * ratio: The Ratio, which shall be adjusted
            * border
        >>> ratio0 = fractions.Fraction(1, 3)
        >>> ratio1 = fractions.Fraction(8, 3)
        >>> border = 2
        >>> JustIntonationPitch._adjust_ratio(ratio0, border)
        fractions.Fraction(4, 3)
        >>> JustIntonationPitch._adjust_ratio(ratio1, border)
        fractions.Fraction(4, 3)

        """

        if border > 1:
            while ratio >= border:
                ratio /= border
            while ratio < 1:
                ratio *= border
        return ratio

    @staticmethod
    def _adjust_float(float_: float, border: int) -> float:
        r"""Multiply float with border, until it is <= 1 and > than border.

        Arguments:
            * float_: The float, which shall be adjusted
            * border
        >>> float0 = 0.5
        >>> float1 = 2
        >>> border = 2
        >>> JustIntonationPitch._adjust_float(float0, border)
        1
        >>> JustIntonationPitch._adjust_float(float1, border)
        1
        """

        if border > 1:
            while float_ > border:
                try:
                    float_ /= border
                except OverflowError:
                    float_ //= border
            while float_ < 1:
                float_ *= border
        return float_

    @staticmethod
    def _adjust_exponent_tuple(
        exponent_tuple: tuple, primes: tuple, border: int
    ) -> tuple:
        r"""Adjust a exponent_tuple and its primes depending on the border.

        Arguments:
            * exponent_tuple: The exponent_tuple, which shall be adjusted
            * primes: Its corresponding primes
            * border
        >>> exponent_tuple0 = (1,)
        >>> primes0 = (3,)
        >>> border = 2
        >>> JustIntonationPitch._adjust_exponent_tuple(exponent_tuple0, primes0, border)
        ((-1, 1), (2, 3))

        """  # TODO(DOCSTRING) Make proper description what actually happens

        if exponent_tuple:
            if border > 1:
                multiplied = functools.reduce(
                    operator.mul, (p ** e for p, e in zip(primes, exponent_tuple))
                )
                res = math.log(border / multiplied, border)
                if res < 0:
                    res -= 1
                res = int(res)
                primes = (border,) + primes
                exponent_tuple = (res,) + exponent_tuple
            return exponent_tuple, primes
        return (1,), (1,)

    @staticmethod
    def _discard_nulls(iterable: typing.Iterable[int]) -> tuple[int, ...]:
        r"""Discard all zeros after the last not 0 - element of an arbitary iterable.

        Return a tuple.
        Arguments:
            * iterable: the iterable, whose 0 - elements shall
              be discarded

        >>> tuple0 = (1, 0, 2, 3, 0, 0, 0)
        >>> ls = [1, 3, 5, 0, 0, 0, 2, 0]
        >>> JustIntonationPitch._discard_nulls(tuple0)
        (1, 0, 2, 3)
        >>> JustIntonationPitch._discard_nulls(ls)
        (1, 3, 5, 0, 0, 0, 2)
        """

        iterable = tuple(iterable)
        c = 0
        for i in reversed(iterable):
            if i != 0:
                break
            c += 1
        if c != 0:
            return iterable[:-c]
        return iterable

    @staticmethod
    def _exponent_tuple_to_pair(exponent_tuple: tuple, primes: tuple) -> tuple:
        r"""Transform a JustIntonationPitch to a (numerator, denominator) - pair.

        Arguments are:
            * JustIntonationPitch -> The exponent_tuple of prime numbers
            * primes -> the referring prime numbers
        >>> myJustIntonationPitch0 = (1, 0, -1)
        >>> myJustIntonationPitch1 = (0, 2, 0)
        >>> myVal0 = (2, 3, 5)
        >>> myVal1 = (3, 5, 7)
        >>> JustIntonationPitch._exponent_tuple_to_pair(myJustIntonationPitch0, myVal0)
        (2, 5)
        >>> JustIntonationPitch._exponent_tuple_to_pair(myJustIntonationPitch0, myVal1)
        (3, 7)
        >>> JustIntonationPitch._exponent_tuple_to_pair(myJustIntonationPitch1, myVal1)
        (25, 1)
        """

        numerator = 1
        denominator = 1
        for number, exponent in zip(primes, exponent_tuple):
            if exponent > 0:
                numerator *= pow(number, exponent)
            elif exponent < 0:
                denominator *= pow(number, -exponent)
        return numerator, denominator

    @staticmethod
    def _exponent_tuple_to_ratio(
        exponent_tuple: tuple, primes: tuple
    ) -> fractions.Fraction:
        r"""Transform a JustIntonationPitch to a fractions.Fraction - Object

        (if installed to a quicktions.fractions.Fraction - Object,
        otherwise to a fractions.fractions.Fraction - Object).

        Arguments are:
            * JustIntonationPitch -> The exponent_tuple of prime numbers
            * primes -> the referring prime numbers for the underlying
                      ._exponent_tuple - Argument (see JustIntonationPitch._exponent_tuple).
        >>> myJustIntonationPitch0 = (1, 0, -1)
        >>> myPrimes= (2, 3, 5)
        >>> JustIntonationPitch._exponent_tuple_to_ratio(myJustIntonationPitch0, myPrimes)
        2/5
        """

        numerator, denominator = JustIntonationPitch._exponent_tuple_to_pair(
            exponent_tuple, primes
        )
        return JustIntonationPitch._adjust_ratio(
            fractions.Fraction(numerator, denominator), 1
        )

    @staticmethod
    def _exponent_tuple_to_float(exponent_tuple: tuple, primes: tuple) -> float:
        r"""Transform a JustIntonationPitch to a float.

        Arguments are:
            * JustIntonationPitch -> The exponent_tuple of prime numbers
            * primes -> the referring prime numbers for the underlying
                      ._exponent_tuple - Argument (see JustIntonationPitch._exponent_tuple).
            * primes-shift -> how many prime numbers shall be skipped
                            (see JustIntonationPitch.primes_shift)

        >>> myJustIntonationPitch0 = (1, 0, -1)
        >>> myJustIntonationPitch1 = (0, 2, 0)
        >>> myPrimes = (2, 3, 5)
        >>> JustIntonationPitch._exponent_tuple_to_float(myJustIntonationPitch0, myPrimes)
        0.4
        """

        numerator, denominator = JustIntonationPitch._exponent_tuple_to_pair(
            exponent_tuple, primes
        )
        try:
            return numerator / denominator
        except OverflowError:
            return numerator // denominator

    @staticmethod
    def _ratio_to_exponent_tuple(ratio: fractions.Fraction) -> tuple:
        r"""Transform a fractions.Fraction - Object to a vector of exponent_tuple.

        :param ratio: The fractions.Fraction, which shall be transformed

        **Example:**

        >>> try:
        >>>     from quicktions import fractions.Fraction
        >>> except ImportError:
        >>>     from fractions import fractions.Fraction
        >>> myRatio0 = fractions.Fraction(3, 2)
        >>> JustIntonationPitch._ratio_to_exponent_tuple(myRatio0)
        (-1, 1)
        """

        factorised_numerator = prime_factors.factors(ratio.numerator)
        factorised_denominator = prime_factors.factors(ratio.denominator)

        factorised_num = prime_factors.factorise(ratio.numerator)
        factorised_den = prime_factors.factorise(ratio.denominator)

        biggest_prime = max(factorised_num + factorised_den)
        exponent_tuple = [0] * primesieve.count_primes(biggest_prime)

        for prime, fac in factorised_numerator:
            if prime > 1:
                exponent_tuple[primesieve.count_primes(prime) - 1] += fac

        for prime, fac in factorised_denominator:
            if prime > 1:
                exponent_tuple[primesieve.count_primes(prime) - 1] -= fac

        return tuple(exponent_tuple)

    @staticmethod
    def _indigestibility(num: int) -> float:
        """Calculate _indigestibility of a number

        The implementation follows Clarence Barlows definition
        given in 'The Ratio Book' (1992).
        Arguments:
            * num -> integer, whose _indigestibility value shall be calculated

        **Example:**

        >>> JustIntonationPitch._indigestibility(1)
        0
        >>> JustIntonationPitch._indigestibility(2)
        1
        >>> JustIntonationPitch._indigestibility(3)
        2.6666666666666665
        """

        decomposed = prime_factors.factorise(num)
        return JustIntonationPitch._indigestibility_of_factorised(decomposed)

    @staticmethod
    def _indigestibility_of_factorised(decomposed):
        decomposed = collections.Counter(decomposed)
        decomposed = zip(decomposed.values(), decomposed.keys())
        summed = ((power * pow(prime - 1, 2)) / prime for power, prime in decomposed)
        return 2 * sum(summed)

    @staticmethod
    def _count_accidentals(accidentals: str) -> int:
        accidental_counter = collections.Counter({"f": 0, "s": 0})
        accidental_counter.update(accidentals)
        for accidental in accidentals:
            if accidental not in ("f", "s"):
                message = "Found unknown accidental '{}' which will be ignored".format(
                    accidental
                )
                warnings.warn(message)
        return (1 * accidental_counter["s"]) - (1 * accidental_counter["f"])

    @staticmethod
    def _get_accidentals(n_accidentals: int) -> str:
        if n_accidentals > 0:
            return "s" * n_accidentals
        else:
            return "f" * abs(n_accidentals)

    # ###################################################################### #
    #                            private methods                             #
    # ###################################################################### #

    def _translate_ratio_or_fractions_argument_to_exponent_tuple(
        self,
        ratio_or_exponent_tuple: typing.Union[
            str, fractions.Fraction, typing.Iterable[int]
        ],
    ) -> tuple[int, ...]:
        if isinstance(ratio_or_exponent_tuple, str):
            numerator, denominator = ratio_or_exponent_tuple.split("/")
            exponent_tuple = self._ratio_to_exponent_tuple(
                fractions.Fraction(int(numerator), int(denominator))
            )
        elif isinstance(ratio_or_exponent_tuple, typing.Iterable):
            exponent_tuple = tuple(ratio_or_exponent_tuple)
        elif hasattr(ratio_or_exponent_tuple, "numerator") and hasattr(
            ratio_or_exponent_tuple, "denominator"
        ):
            exponent_tuple = self._ratio_to_exponent_tuple(
                fractions.Fraction(
                    ratio_or_exponent_tuple.numerator,
                    ratio_or_exponent_tuple.denominator,
                )
            )
        else:
            message = "Unknown type '{}' of object '{}' for 'ratio_or_exponent_tuple' argument.".format(
                type(ratio_or_exponent_tuple), ratio_or_exponent_tuple
            )
            raise NotImplementedError(message)
        return exponent_tuple

    @decorators.add_copy_option
    def _math(  # type: ignore
        self, other: JustIntonationPitch, operation: typing.Callable
    ) -> JustIntonationPitch:
        exponent_tuple0, exponent_tuple1 = JustIntonationPitch._adjust_exponent_lengths(
            self.exponent_tuple, other.exponent_tuple
        )
        self.exponent_tuple = tuple(
            operation(x, y) for x, y in zip(exponent_tuple0, exponent_tuple1)
        )

    # ###################################################################### #
    #                            magic methods                               #
    # ###################################################################### #

    def __float__(self) -> float:
        """Return the float of a JustIntonationPitch - object.

        These are the same:
            float(myJustIntonationPitch.ratio) == float(myJustIntonationPitch).
        Note the difference that the second version might be slightly
        more performant.

        **Example:**

        >>> jip0 = JustIntonationPitch((-1, 1))
        >>> float(jip0)
        1.5
        >>> float(jip0.ratio)
        1.5
        """

        return self._exponent_tuple_to_float(self.exponent_tuple, self.prime_tuple)

    def __repr__(self) -> str:
        ratio = str(self.ratio)
        if len(ratio) == 1:
            ratio += "/1"
        return f"JustIntonationPitch('{ratio}')"

    def __abs__(self):
        if self.numerator > self.denominator:
            return copy.deepcopy(self)
        else:
            exponent_tuple = tuple(-v for v in iter(self.exponent_tuple))
            return type(self)(exponent_tuple, self.concert_pitch)

    # ###################################################################### #
    #                            properties                                  #
    # ###################################################################### #

    @property
    def exponent_tuple(self) -> tuple:
        return self._exponent_tuple

    @exponent_tuple.setter
    def exponent_tuple(
        self,
        exponent_tuple: typing.Iterable[int],
    ) -> None:
        self._exponent_tuple = self._discard_nulls(exponent_tuple)

    @property
    def prime_tuple(self) -> tuple:
        r"""Return ascending list of primes, until the highest contained Prime.

        **Example:**

        >>> jip0 = JustIntonationPitch((0, 1, 2))
        >>> jip0.exponent_tuple
        (2, 3, 5)
        >>> jip1 = JustIntonationPitch((0, -1, 0, 0, 1), 1)
        >>> jip1.exponent_tuple
        (2, 3, 5, 7, 11)
        """

        return tuple(primesieve_numpy.n_primes(len(self.exponent_tuple)))

    @property
    def occupied_primes(self) -> tuple:
        """Return all occurring prime numbers of a JustIntonationPitch object."""

        return tuple(
            prime
            for prime, exponent in zip(self.prime_tuple, self.exponent_tuple)
            if exponent != 0
        )

    @property
    def concert_pitch(self) -> parameters.abc.Pitch:
        return self._concert_pitch

    @concert_pitch.setter
    def concert_pitch(self, concert_pitch: ConcertPitch) -> None:
        if not isinstance(concert_pitch, parameters.abc.Pitch):
            concert_pitch = DirectPitch(concert_pitch)

        self._concert_pitch = concert_pitch

    @property
    def frequency(self) -> float:
        return float(self.ratio * self.concert_pitch.frequency)

    @property
    def ratio(self) -> fractions.Fraction:
        """Return the JustIntonationPitch transformed to a Ratio.

        **Example:**

        >>> jip0 = JustIntonationPitch((0, 0, 1,))
        >>> jip0.ratio
        fractions.Fraction(5, 4)
        >>> jip0 = JustIntonationPitch("3/2")
        >>> jip0.ratio
        fractions.Fraction(3, 2)
        """

        return JustIntonationPitch._exponent_tuple_to_ratio(
            self.exponent_tuple, self.prime_tuple
        )

    @property
    def numerator(self) -> int:
        """Return the numerator of a JustIntonationPitch - object.

        **Example:**

        >>> jip0 = JustIntonationPitch((0, -1,))
        >>> jip0.numerator
        1
        """

        numerator = 1
        for number, exponent in zip(self.prime_tuple, self.exponent_tuple):
            if exponent > 0:
                numerator *= pow(number, exponent)
        return numerator

    @property
    def denominator(self) -> int:
        """Return the denominator of :class:`JustIntonationPitch`.

        **Example:**

        >>> jip0 = JustIntonationPitch((0, 1,))
        >>> jip0.denominator
        1
        """

        denominator = 1
        for number, exponent in zip(self.prime_tuple, self.exponent_tuple):
            if exponent < 0:
                denominator *= pow(number, -exponent)
        return denominator

    @property
    def cents(self) -> float:
        return self.ratio_to_cents(self.ratio)

    @property
    def factorised(self) -> tuple:
        """Return factorised / decomposed version of itsef.

        **Example:**

        >>> jip0 = JustIntonationPitch((0, 0, 1,))
        >>> jip0.factorised
        (2, 2, 5)
        >>> jip1 = JustIntonationPitch("7/6")
        >>> jip1.factorised
        (2, 3, 7)
        """

        exponent_tuple = self.exponent_tuple
        prime_tuple = self.prime_tuple
        exponent_tuple_adjusted, primes_adjusted = type(self)._adjust_exponent_tuple(
            exponent_tuple, prime_tuple, 1
        )
        decomposed = (
            [p] * abs(e) for p, e in zip(primes_adjusted, exponent_tuple_adjusted)
        )
        return tuple(functools.reduce(operator.add, decomposed))

    @property
    def factorised_numerator_and_denominator(self) -> tuple:
        exponent_tuple = self.exponent_tuple
        prime_tuple = self.prime_tuple
        exponent_tuple_adjusted, prime_tuple_adjusted = type(
            self
        )._adjust_exponent_tuple(exponent_tuple, prime_tuple, 1)
        numerator_denominator: list[list[list[int]]] = [[[]], [[]]]
        for prime, exponent in zip(prime_tuple_adjusted, exponent_tuple_adjusted):
            if exponent > 0:
                idx = 0
            else:
                idx = 1
            numerator_denominator[idx].append([prime] * abs(exponent))
        return tuple(
            functools.reduce(operator.add, decomposed)
            for decomposed in numerator_denominator
        )

    @property
    def octave(self) -> int:
        ct = self.cents
        ref, exp = 1200, 0
        while ref * exp <= ct:
            exp += 1
        while ref * exp > ct:
            exp -= 1
        return exp

    @property
    def helmholtz_ellis_just_intonation_notation_commas(
        self,
    ) -> parameters.commas.CommaCompound:
        """Commas of JustIntonationPitch."""

        prime_to_exponent_dict = {
            prime: exponent
            for prime, exponent in zip(self.prime_tuple, self.exponent_tuple)
            if exponent != 0 and prime not in (2, 3)
        }
        return parameters.commas.CommaCompound(
            prime_to_exponent_dict,
            parameters.pitches_constants.DEFAULT_PRIME_TO_COMMA_DICT,
        )

    @property
    def closest_pythagorean_interval(self) -> JustIntonationPitch:
        if len(self.helmholtz_ellis_just_intonation_notation_commas) > 0:
            closest_pythagorean_interval = self - type(self)(
                functools.reduce(
                    operator.mul, self.helmholtz_ellis_just_intonation_notation_commas
                )
            )
            closest_pythagorean_interval.normalize()
        else:
            closest_pythagorean_interval = self.normalize(mutate=False)  # type: ignore

        return closest_pythagorean_interval

    @property
    def cent_deviation_from_closest_western_pitch_class(self) -> float:
        deviation_by_helmholtz_ellis_just_intonation_notation_commas = (
            JustIntonationPitch(
                self.helmholtz_ellis_just_intonation_notation_commas.ratio
            ).cents
        )
        closest_pythagorean_interval = self.closest_pythagorean_interval
        if len(closest_pythagorean_interval.exponent_tuple) >= 2:
            pythagorean_deviation = self.closest_pythagorean_interval.exponent_tuple[
                1
            ] * (JustIntonationPitch("3/2").cents - 700)
        else:
            pythagorean_deviation = 0
        return (
            deviation_by_helmholtz_ellis_just_intonation_notation_commas
            + pythagorean_deviation
        )

    @property
    def blueprint(  # type: ignore
        self, ignore: typing.Sequence[int] = (2,)
    ) -> tuple[tuple[int, ...], ...]:
        blueprint = []
        for factorised in self.factorised_numerator_and_denominator:
            factorised = tuple(fac for fac in factorised if fac not in ignore)
            counter = collections.Counter(collections.Counter(factorised).values())
            if counter:
                maxima = max(counter.keys())
                blueprint.append(tuple(counter[idx + 1] for idx in range(maxima)))
            else:
                blueprint.append(tuple([]))
        return tuple(blueprint)

    @property
    def tonality(self) -> bool:
        """Return the tonality (bool) of a JustIntonationPitch - object.

        The tonality of a JustIntonationPitch   - may be True (otonality) if
        the exponent of the highest occurring prime number is a
        positive number and False if the exponent is a
        negative number (utonality).

        **Example:**

        >>> jip0 = JustIntonationPitch((-2. 1))
        >>> jip0.tonality
        True
        >>> jip1 = JustIntonationPitch((-2, -1))
        >>> jip1.tonality
        False
        >>> jip2 = JustIntonationPitch([])
        >>> jip2.tonality
        True
        """

        if self.exponent_tuple:
            maxima = max(self.exponent_tuple)
            minima = min(self.exponent_tuple)
            test = (
                maxima <= 0 and minima < 0,
                minima < 0
                and self.exponent_tuple.index(minima)
                > self.exponent_tuple.index(maxima),
            )
            if any(test):
                return False

        return True

    @property
    def harmonic(self) -> int:
        """Return the nth - harmonic / subharmonic the pitch may represent.

        :return: May be positive for harmonic and negative for
            subharmonic pitches. If the return - value is 0,
            the interval may occur neither between the first harmonic
            and any other pitch of the harmonic scale nor
            between the first subharmonic in the and any other
            pitch of the subharmonic scale.

        **Example:**

        >>> jip0 = JustIntonationPitch((0, 1))
        >>> jip0.ratio
        fractions.Fraction(3, 2)
        >>> jip0.harmonic
        3
        >>> jip1 = JustIntonationPitch((-1,), 2)
        >>> jip1.harmonic
        -3
        """

        ratio = self.ratio

        if ratio.denominator % 2 == 0:
            return ratio.numerator
        elif ratio.numerator % 2 == 0:
            return -ratio.denominator
        elif ratio == fractions.Fraction(1, 1):
            return 1
        else:
            return 0

    @property
    def primes_for_numerator_and_denominator(self) -> tuple:
        return tuple(
            tuple(sorted(set(prime_factors.factorise(n))))
            for n in (self.numerator, self.denominator)
        )

    @property
    def harmonicity_wilson(self) -> int:
        decomposed = self.factorised
        return int(sum(filter(lambda x: x != 2, decomposed)))

    @property
    def harmonicity_vogel(self) -> int:
        decomposed = self.factorised
        decomposed_filtered = tuple(filter(lambda x: x != 2, decomposed))
        am_2 = len(decomposed) - len(decomposed_filtered)
        return int(sum(decomposed_filtered) + am_2)

    @property
    def harmonicity_euler(self) -> int:
        """Return the 'gradus suavitatis' of euler.

        A higher number means a less consonant interval /
        a more complicated harmony.
        euler(1/1) is definied as 1.

        **Example:**

        >>> jip0 = JustIntonationPitch((0, 1,))
        >>> jip1 = JustIntonationPitch()
        >>> jip2 = JustIntonationPitch((0, 0, 1,))
        >>> jip3 = JustIntonationPitch((0, 0, -1,))
        >>> jip0.harmonicity_euler
        4
        >>> jip1.harmonicity_euler
        1
        >>> jip2.harmonicity_euler
        7
        >>> jip3.harmonicity_euler
        8
        """

        decomposed = self.factorised
        return 1 + sum(x - 1 for x in decomposed)

    @property
    def harmonicity_barlow(self) -> float:
        r"""Calculate the barlow-harmonicity of an interval.

        This implementation follows Clarence Barlows definition, given
        in 'The Ratio Book' (1992).

        A higher number means a more harmonic interval / a less
        complex harmony.

        barlow(1/1) is definied as infinite.

        **Example:**

        >>> jip0 = JustIntonationPitch((0, 1,))
        >>> jip1 = JustIntonationPitch()
        >>> jip2 = JustIntonationPitch((0, 0, 1,))
        >>> jip3 = JustIntonationPitch((0, 0, -1,))
        >>> jip0.harmonicity_barlow
        0.27272727272727276
        >>> jip1.harmonicity_barlow # 1/1 is infinite harmonic
        inf
        >>> jip2.harmonicity_barlow
        0.11904761904761904
        >>> jip3.harmonicity_barlow
        -0.10638297872340426
        """

        def sign(x):
            return (1, -1)[x < 0]

        numerator_denominator_decomposed = self.factorised_numerator_and_denominator
        indigestibility_numerator = JustIntonationPitch._indigestibility_of_factorised(
            numerator_denominator_decomposed[0]
        )
        indigestibility_denominator = (
            JustIntonationPitch._indigestibility_of_factorised(
                numerator_denominator_decomposed[1]
            )
        )
        if indigestibility_numerator == 0 and indigestibility_denominator == 0:
            return float("inf")
        return sign(indigestibility_numerator - indigestibility_denominator) / (
            indigestibility_numerator + indigestibility_denominator
        )

    @property
    def harmonicity_simplified_barlow(self) -> float:
        r"""Calculate a simplified barlow-harmonicity of an interval.

        This implementation follows Clarence Barlows definition, given
        in 'The Ratio Book' (1992), with the difference that
        only positive numbers are returned and that (1/1) is
        defined as 1 instead of infinite.

        >>> jip0 = JustIntonationPitch((0, 1,))
        >>> jip1 = JustIntonationPitch()
        >>> jip2 = JustIntonationPitch((0, 0, 1,))
        >>> jip3 = JustIntonationPitch((0, 0, -1,))
        >>> jip0.harmonicity_simplified_barlow
        0.27272727272727276
        >>> jip1.harmonicity_simplified_barlow # 1/1 is not infinite but 1
        1
        >>> jip2.harmonicity_simplified_barlow
        0.11904761904761904
        >>> jip3.harmonicity_simplified_barlow # positive return value
        0.10638297872340426
        """

        barlow = abs(self.harmonicity_barlow)
        if barlow == float("inf"):
            return 1
        return barlow

    @property
    def harmonicity_tenney(self) -> float:
        r"""Calculate Tenneys harmonic distance of an interval

        A higher number
        means a more consonant interval / a less
        complicated harmony.

        tenney(1/1) is definied as 0.

        >>> jip0 = JustIntonationPitch((0, 1,))
        >>> jip1 = JustIntonationPitch()
        >>> jip2 = JustIntonationPitch((0, 0, 1,))
        >>> jip3 = JustIntonationPitch((0, 0, -1,))
        >>> jip0.harmonicity_tenney
        2.584962500721156
        >>> jip1.harmonicity_tenney
        0.0
        >>> jip2.harmonicity_tenney
        4.321928094887363
        >>> jip3.harmonicity_tenney
        -0.10638297872340426
        """

        ratio = self.ratio
        num = ratio.numerator
        de = ratio.denominator
        return math.log(num * de, 2)

    @property
    def level(self) -> int:
        if self.prime_tuple:
            return abs(
                functools.reduce(
                    math.gcd, tuple(filter(lambda x: x != 0, self.exponent_tuple))  # type: ignore
                )
            )
        else:
            return 1

    # ###################################################################### #
    #                            public methods                              #
    # ###################################################################### #

    def get_closest_pythagorean_pitch_name(self, reference: str = "a") -> str:
        """"""

        # TODO(for future usage: type reference as typing.Literal[] instead of str)

        # TODO(split method, make it more readable)

        # TODO(Add documentation)

        diatonic_pitch_name, accidentals = reference[0], reference[1:]
        n_accidentals_in_reference = JustIntonationPitch._count_accidentals(accidentals)
        position_of_diatonic_pitch_in_cycle_of_fifths = (
            parameters.pitches_constants.DIATONIC_PITCH_NAME_CYCLE_OF_FIFTH_TUPLE.index(
                diatonic_pitch_name
            )
        )

        closest_pythagorean_interval = self.closest_pythagorean_interval
        try:
            n_fifths = closest_pythagorean_interval.exponent_tuple[1]
        # for 1/1
        except IndexError:
            n_fifths = 0

        # 1. Find new diatonic pitch name
        n_steps_in_diatonic_pitch_name = n_fifths % 7
        nth_diatonic_pitch = (
            position_of_diatonic_pitch_in_cycle_of_fifths
            + n_steps_in_diatonic_pitch_name
        ) % 7
        new_diatonic_pitch = (
            parameters.pitches_constants.DIATONIC_PITCH_NAME_CYCLE_OF_FIFTH_TUPLE[
                nth_diatonic_pitch
            ]
        )

        # 2. Find new accidentals
        n_accidentals_in_closest_pythagorean_pitch = (
            (position_of_diatonic_pitch_in_cycle_of_fifths + n_fifths) // 7
        ) + n_accidentals_in_reference
        new_accidentals = JustIntonationPitch._get_accidentals(
            n_accidentals_in_closest_pythagorean_pitch
        )

        return "".join((new_diatonic_pitch, new_accidentals))

    @decorators.add_copy_option
    def register(self, octave: int) -> JustIntonationPitch:  # type: ignore
        """Move :class:`JustIntonationPitch` to the given octave.

        :param octave: 0 for the octave from 1/1 to 2/1, negative values for octaves
            below 1/1 and positive values for octaves above 2/1.
        :type octave: int

        **Example:**

        >>> from mutwo.parameters import pitches
        >>> p = pitches.JustIntonationPitch('3/2')
        >>> p.register(1)
        >>> p
        JustIntonationPitch(6/2)
        >>> p.register(-1)
        >>> p
        JustIntonationPitch(3/4)
        >>> p.register(0)
        >>> p
        JustIntonationPitch(3/2)
        """

        normalized_just_intonation_pitch = self.normalize(mutate=False)  # type: ignore
        factor = 2 ** abs(octave)
        if octave < 1:
            added = type(self)(fractions.Fraction(1, factor))
        else:
            added = type(self)(fractions.Fraction(factor, 1))
        self.exponent_tuple = (normalized_just_intonation_pitch + added).exponent_tuple  # type: ignore

    @decorators.add_copy_option
    def move_to_closest_register(  # type: ignore
        self, reference: JustIntonationPitch
    ) -> JustIntonationPitch:
        reference_register = reference.octave

        best = None
        for adaption in range(-1, 2):
            candidate: JustIntonationPitch = self.register(reference_register + adaption, mutate=False)  # type: ignore
            difference = abs((candidate - reference).cents)
            set_best = True
            if best and difference > best[1]:
                set_best = False

            if set_best:
                best = (candidate, difference)

        if best:
            self.exponent_tuple = best[0].exponent_tuple
        else:
            raise NotImplementedError(
                f"Couldn't find closest register of '{self}' to '{reference}'."
            )

    @decorators.add_copy_option
    def normalize(self, prime: int = 2) -> JustIntonationPitch:  # type: ignore
        """Normalize :class:`JustIntonationPitch`.

        :param prime: The normalization period (2 for octave,
            3 for twelfth, ...). Default to 2.
        :type prime: int

        **Example:**

        >>> from mutwo.parameters import pitches
        >>> p = pitches.JustIntonationPitch('12/2')
        >>> p.normalize()
        >>> p
        JustIntonationPitch(3/2)
        """
        ratio = self.ratio
        adjusted = type(self)._adjust_ratio(ratio, prime)
        self.exponent_tuple = (
            self._translate_ratio_or_fractions_argument_to_exponent_tuple(adjusted)
        )

    @decorators.add_copy_option
    def inverse(  # type: ignore
        self, axis: typing.Optional[JustIntonationPitch] = None
    ) -> JustIntonationPitch:
        """Inverse current pitch on given axis.

        :param axis: The :class:`JustIntonationPitch` from which the
            pitch shall be inversed.
        :type axis: JustIntonationPitch, optional

        **Example:**

        >>> from mutwo.parameters import pitches
        >>> p = pitches.JustIntonationPitch('3/2')
        >>> p.inverse()
        >>> p
        JustIntonationPitch(2/3)
        """

        if axis is None:
            exponent_tuple = tuple(map(lambda x: -x, self.exponent_tuple))
        else:
            distance = self - axis
            exponent_tuple = (axis - distance).exponent_tuple
        self.exponent_tuple = exponent_tuple

    @decorators.add_copy_option
    def add(self, pitch_interval: parameters.abc.PitchInterval) -> JustIntonationPitch:
        """Add :class:`JustIntonationPitch` to current pitch.

        :param other: The :class:`JustIntonationPitch` to add to
            the current pitch.

        **Example:**

        >>> from mutwo.parameters import pitches
        >>> p = pitches.JustIntonationPitch('3/2')
        >>> p.add(pitches.JustIntonationPitch('3/2'))
        >>> p
        JustIntonationPitch(9/4)
        """
        if isinstance(pitch_interval, JustIntonationPitch):
            self._math(pitch_interval, operator.add)
        else:
            self.exponent_tuple = self._ratio_to_exponent_tuple(
                self.ratio * self.cents_to_ratio(pitch_interval.cents)
            )
        return self

    @decorators.add_copy_option
    def subtract(
        self, pitch_interval: parameters.abc.PitchInterval
    ) -> JustIntonationPitch:
        """Subtract :class:`JustIntonationPitch` from current pitch.

        :param other: The :class:`JustIntonationPitch` to subtract from
            the current pitch.

        **Example:**

        >>> from mutwo.parameters import pitches
        >>> p = pitches.JustIntonationPitch('9/4')
        >>> p.subtract(pitches.JustIntonationPitch('3/2'))
        >>> p
        JustIntonationPitch(3/2)
        """

        if isinstance(pitch_interval, JustIntonationPitch):
            self._math(pitch_interval, operator.sub)
        else:
            self.exponent_tuple = self._ratio_to_exponent_tuple(
                self.ratio / self.cents_to_ratio(pitch_interval.cents)
            )
        return self

    @decorators.add_copy_option
    def intersection(
        self, other: JustIntonationPitch, strict: bool = False
    ) -> JustIntonationPitch:
        """Make intersection with other :class:`JustIntonationPitch`.

        :param other: The :class:`JustIntonationPitch` to build the
            intersection with.
        :param strict: If set to ``True`` only exponent_tuple are included
            into the intersection if their value is equal. If set to
            ``False`` the method will also include exponent_tuple if both
            pitches own them on the same axis but with different values
            (the method will take the smaller exponent).
        :type strict: bool

        **Example:**

        >>> from mutwo.parameters import pitches
        >>> p0 = pitches.JustIntonationPitch('5/3')
        >>> p0.intersection(pitches.JustIntonationPitch('7/6'))
        >>> p0
        JustIntonationPitch(1/3)
        >>> p1 = pitches.JustIntonationPitch('9/7')
        >>> p1.intersection(pitches.JustIntonationPitch('3/2'))
        >>> p1
        JustIntonationPitch(3/1)
        >>> p2 = pitches.JustIntonationPitch('9/7')
        >>> p2.intersection(pitches.JustIntonationPitch('3/2'), strict=True)
        >>> p2
        JustIntonationPitch(1/1)
        """

        def is_negative(number: int):
            return number < 0

        def intersect_exponent_tuple(exponent_tuple: tuple[int, int]) -> int:
            intersected_exponent = 0

            if strict:
                test_for_valid_strict_mode = exponent_tuple[0] == exponent_tuple[1]
            else:
                test_for_valid_strict_mode = True

            if 0 not in exponent_tuple and test_for_valid_strict_mode:
                are_negative = (is_negative(exponent) for exponent in exponent_tuple)
                all_are_negative = all(are_negative)
                all_are_positive = all(not boolean for boolean in are_negative)

                if all_are_negative:
                    intersected_exponent = max(exponent_tuple)
                elif all_are_positive:
                    intersected_exponent = min(exponent_tuple)

            return intersected_exponent

        intersected_exponent_tuple = tuple(
            map(
                intersect_exponent_tuple, zip(self.exponent_tuple, other.exponent_tuple)
            )
        )
        self.exponent_tuple = intersected_exponent_tuple


class Partial(object):
    """Abstract representation of a harmonic spectrum partial.

    :param nth_partial: The number of the partial (starting with 1
        for the root note).
    :type nth_partial: int
    :param tonality: ``True`` for overtone and ``False`` for a (theoretical)
        undertone. Default to ``True``.
    :type tonality: bool

    **Example:**

    >>> from mutwo.parameters import pitches
    >>> strong_clarinet_partials = (
        pitches.Partial(1),
        pitches.Partial(3),
        pitches.Partial(5),
        pitches.Partial(7),
    )
    """

    def __init__(self, nth_partial: int, tonality: bool = True):
        self._nth_partial = nth_partial
        self._tonality = tonality

    @property
    def nth_partial(self) -> int:
        return self._nth_partial

    @property
    def tonality(self) -> int:
        return self._tonality

    def __repr__(self) -> str:
        return f"Partial({self.nth_partial}, {self.tonality})"


class CommonHarmonic(JustIntonationPitch):
    """:class:`JustIntonationPitch` which is the common harmonic between two or more other pitches.

    :param partials: Tuple which contains partial numbers.
    :type partials: tuple[Partial, ...]
    :param ratio_or_exponent_tuple: see the documentation of :class:`JustIntonationPitch`
    :type ratio_or_exponent_tuple: typing.Union[str, fractions.Fraction, typing.Iterable[int]]
    :param concert_pitch: see the documentation of :class:`JustIntonationPitch`
    :type concert_pitch: typing.Union[constants.Real, parameters.abc.Pitch]
    """

    def __init__(
        self,
        partial_tuple: tuple[Partial, ...],
        ratio_or_exponent_tuple: typing.Union[
            str, fractions.Fraction, typing.Iterable[int]
        ] = "1/1",
        concert_pitch: ConcertPitch = None,
        *args,
        **kwargs,
    ):
        super().__init__(ratio_or_exponent_tuple, concert_pitch, *args, **kwargs)
        self.partial_tuple = partial_tuple

    def __repr__(self) -> str:
        return f"CommonHarmonic({self.ratio}, {self.partial_tuple})"


class EqualDividedOctavePitch(parameters.abc.Pitch):
    """Pitch that is tuned to an Equal divided octave tuning system.

    :param n_pitch_classes_per_octave: how many pitch classes in each octave
        occur (for instance 12 for a chromatic system, 24 for quartertones, etc.)
    :param pitch_class: The pitch class of the new :class:`EqualDividedOctavePitch` object.
    :param octave: The octave of the new :class:`EqualDividedOctavePitch` object (where 0 is
        the middle octave, 1 is one octave higher and -1 is one octave lower).
    :param concert_pitch_pitch_class: The pitch class of the reference pitch (for
        instance 9 in a chromatic 12 tone system where `a` should be the reference
        pitch).
    :param concert_pitch_octave: The octave of the reference pitch.
    :param concert_pitch: The frequency of the reference pitch (for instance 440 for a).

    >>> from mutwo.parameters import pitches
    >>> # making a middle `a`
    >>> pitches.EqualDividedOctavePitch(12, 9, 4, 9, 4, 440)
    """

    def __init__(
        self,
        n_pitch_classes_per_octave: int,
        pitch_class: constants.Real,
        octave: int,
        concert_pitch_pitch_class: constants.Real,
        concert_pitch_octave: int,
        concert_pitch: ConcertPitch = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            *args,
            **kwargs,
        )
        if concert_pitch is None:
            concert_pitch = parameters.pitches_constants.DEFAULT_CONCERT_PITCH

        self._n_pitch_classes_per_octave = n_pitch_classes_per_octave
        self.pitch_class = pitch_class
        self.octave = octave
        self.concert_pitch_pitch_class = concert_pitch_pitch_class
        self.concert_pitch_octave = concert_pitch_octave
        self.concert_pitch = concert_pitch  # type: ignore

    def _assert_correct_pitch_class(self, pitch_class: constants.Real) -> None:
        """Makes sure the respective pitch_class is within the allowed range."""

        try:
            assert all(
                (pitch_class <= self.n_pitch_classes_per_octave - 1, pitch_class >= 0)
            )
        except AssertionError:
            message = (
                "Invalid pitch class {}!. Pitch_class has to be in range (min = 0, max"
                " = {}).".format(pitch_class, self.n_pitch_classes_per_octave - 1)
            )
            raise ValueError(message)

    def _fetch_n_pitch_classes_difference(
        self,
        pitch_interval_or_n_pitch_classes_difference: typing.Union[
            parameters.abc.PitchInterval, constants.Real
        ],
    ) -> constants.Real:
        if isinstance(
            pitch_interval_or_n_pitch_classes_difference, parameters.abc.PitchInterval
        ):
            return (
                pitch_interval_or_n_pitch_classes_difference.cents
                / self.n_cents_per_step
            )
        else:
            return pitch_interval_or_n_pitch_classes_difference

    @property
    def n_pitch_classes_per_octave(self) -> int:
        """Defines in how many different pitch classes one octave get divided."""
        return self._n_pitch_classes_per_octave

    @property
    def concert_pitch(self) -> parameters.abc.Pitch:
        """The referential concert pitch for the respective pitch object."""
        return self._concert_pitch

    @concert_pitch.setter
    def concert_pitch(self, concert_pitch: ConcertPitch) -> None:
        if not isinstance(concert_pitch, parameters.abc.Pitch):
            concert_pitch = DirectPitch(concert_pitch)

        self._concert_pitch = concert_pitch

    @property
    def concert_pitch_pitch_class(self) -> constants.Real:
        """The pitch class of the referential concert pitch."""
        return self._concert_pitch_pitch_class

    @concert_pitch_pitch_class.setter
    def concert_pitch_pitch_class(self, pitch_class: constants.Real) -> None:
        self._assert_correct_pitch_class(pitch_class)
        self._concert_pitch_pitch_class = pitch_class

    @property
    def pitch_class(self) -> constants.Real:
        """The pitch class of the pitch."""
        return self._pitch_class

    @pitch_class.setter
    def pitch_class(self, pitch_class: constants.Real) -> None:
        self._assert_correct_pitch_class(pitch_class)
        self._pitch_class = pitch_class

    @property
    def step_factor(self):
        """The factor with which to multiply a frequency to reach the next pitch."""
        return pow(2, 1 / self.n_pitch_classes_per_octave)

    @property
    def n_cents_per_step(self) -> float:
        """This property describes how many cents are between two adjacent pitches."""
        return self.ratio_to_cents(self.step_factor)

    @property
    def frequency(self) -> float:
        n_octaves_distant_to_concert_pitch = self.octave - self.concert_pitch_octave
        n_pitch_classes_distant_to_concert_pitch = (
            self.pitch_class - self.concert_pitch_pitch_class
        )
        distance_to_concert_pitch_in_cents = (
            n_octaves_distant_to_concert_pitch * 1200
        ) + (self.n_cents_per_step * n_pitch_classes_distant_to_concert_pitch)
        distance_to_concert_pitch_as_factor = self.cents_to_ratio(
            distance_to_concert_pitch_in_cents
        )
        return float(self.concert_pitch.frequency * distance_to_concert_pitch_as_factor)

    def __sub__(self, other: EqualDividedOctavePitch) -> constants.Real:
        """Calculates the interval between two ``EqualDividedOctave`` pitches."""

        try:
            assert self.n_pitch_classes_per_octave == other.n_pitch_classes_per_octave
        except AssertionError:
            message = (
                "Can't calculate the interval between to different"
                " EqualDividedOctavePitch objects with different value for"
                " 'n_pitch_classes_per_octave'."
            )
            raise ValueError(message)

        n_pitch_classes_difference = self.pitch_class - other.pitch_class
        n_octaves_difference = self.octave - other.octave
        return n_pitch_classes_difference + (
            n_octaves_difference * self.n_pitch_classes_per_octave
        )

    def _math(
        self,
        n_pitch_classes_difference: constants.Real,
        operator: typing.Callable[[constants.Real, constants.Real], constants.Real],
    ) -> None:
        new_pitch_class = operator(self.pitch_class, n_pitch_classes_difference)
        n_octaves_difference = new_pitch_class // self.n_pitch_classes_per_octave
        new_pitch_class = new_pitch_class % self.n_pitch_classes_per_octave
        new_octave = self.octave + n_octaves_difference
        self.pitch_class = new_pitch_class
        self.octave = int(new_octave)

    @decorators.add_copy_option
    def add(  # type: ignore
        self,
        pitch_interval_or_n_pitch_classes_difference: typing.Union[
            parameters.abc.PitchInterval, constants.Real
        ],
    ) -> EqualDividedOctavePitch:  # type: ignore
        """Transposes the ``EqualDividedOctavePitch`` by n_pitch_classes_difference."""

        n_pitch_classes_difference = self._fetch_n_pitch_classes_difference(
            pitch_interval_or_n_pitch_classes_difference
        )
        self._math(n_pitch_classes_difference, operator.add)

    @decorators.add_copy_option
    def subtract(  # type: ignore
        self,
        pitch_interval_or_n_pitch_classes_difference: typing.Union[
            parameters.abc.PitchInterval, constants.Real
        ],
    ) -> EqualDividedOctavePitch:  # type: ignore
        """Transposes the ``EqualDividedOctavePitch`` by n_pitch_classes_difference."""

        n_pitch_classes_difference = self._fetch_n_pitch_classes_difference(
            pitch_interval_or_n_pitch_classes_difference
        )
        self._math(n_pitch_classes_difference, operator.sub)


# TODO(add something similar to scamps SpellingPolicy (don't hard code
# if mutwo shall write a flat or sharp)
# TODO(add translation from octave number to notated octave (4 -> ', 5 -> '', ..))


class WesternPitch(EqualDividedOctavePitch):
    """Pitch with a traditional Western nomenclature.

    :param pitch_class_or_pitch_class_name: Name or number of the pitch class of the
        new ``WesternPitch`` object. The nomenclature is English (c, d, e, f, g, a, b).
        It uses an equal divided octave system in 12 chromatic steps. Accidentals are
        indicated by (s = sharp) and (f = flat). Further microtonal accidentals are
        supported (see
        :const:`mutwo.parameters.pitches_constants.ACCIDENTAL_NAME_TO_PITCH_CLASS_MODIFICATION_DICT`
        for all supported accidentals).
    :param octave: The octave of the new :class:`WesternPitch` object. Indications for the
        specific octave follow the MIDI Standard where 4 is defined as one line.

    >>> from mutwo.parameters import pitches
    >>> pitches.WesternPitch('cs', 4)  # c-sharp 4
    >>> pitches.WesternPitch('aqs', 2)  # a-quarter-sharp 2
    """

    def __init__(
        self,
        pitch_class_or_pitch_class_name: PitchClassOrPitchClassName = 0,
        octave: int = 4,
        concert_pitch_pitch_class: constants.Real = None,
        concert_pitch_octave: int = None,
        concert_pitch: ConcertPitch = None,
        *args,
        **kwargs,
    ):
        if concert_pitch_pitch_class is None:
            concert_pitch_pitch_class = (
                parameters.pitches_constants.DEFAULT_CONCERT_PITCH_PITCH_CLASS_FOR_WESTERN_PITCH
            )

        if concert_pitch_octave is None:
            concert_pitch_octave = (
                parameters.pitches_constants.DEFAULT_CONCERT_PITCH_OCTAVE_FOR_WESTERN_PITCH
            )

        (
            pitch_class,
            pitch_class_name,
        ) = self._translate_pitch_class_or_pitch_class_name_to_pitch_class_and_pitch_class_name(
            pitch_class_or_pitch_class_name
        )
        super().__init__(
            12,
            pitch_class,
            octave,
            concert_pitch_pitch_class,
            concert_pitch_octave,
            concert_pitch,
            *args,
            **kwargs,
        )

        self._pitch_class_name = pitch_class_name

    @staticmethod
    def _translate_pitch_class_or_pitch_class_name_to_pitch_class_and_pitch_class_name(
        pitch_class_or_pitch_class_name: PitchClassOrPitchClassName,
    ) -> tuple:
        """Helper function to initialise a WesternPitch from a number or a string.

        A number has to represent the pitch class while the name has to use
        the Western English nomenclature with the form
        DIATONICPITCHCLASSNAME-ACCIDENTAL (e.g. "cs" for c-sharp,
        "gqf" for g-quarter-flat, "b" for b)
        """
        if isinstance(pitch_class_or_pitch_class_name, numbers.Real):
            pitch_class = float(pitch_class_or_pitch_class_name)
            pitch_class_name = WesternPitch._translate_pitch_class_to_pitch_class_name(
                pitch_class_or_pitch_class_name
            )
        elif isinstance(pitch_class_or_pitch_class_name, str):
            pitch_class = WesternPitch._translate_pitch_class_name_to_pitch_class(
                pitch_class_or_pitch_class_name
            )
            pitch_class_name = pitch_class_or_pitch_class_name
        else:
            message = "Can't initalise pitch_class by '{}' of type '{}'.".format(
                pitch_class_or_pitch_class_name, type(pitch_class_or_pitch_class_name)
            )
            raise TypeError(message)

        return pitch_class, pitch_class_name

    @staticmethod
    def _translate_accidental_to_pitch_class_modifications(
        accidental: str,
    ) -> constants.Real:
        """Helper function to translate an accidental to its pitch class modification.

        Raises an error if the accidental hasn't been defined yet in
        mutwo.parameters.parameters.parameters.pitches_constants.ACCIDENTAL_NAME_TO_PITCH_CLASS_MODIFICATION_DICT.
        """
        try:
            return parameters.pitches_constants.ACCIDENTAL_NAME_TO_PITCH_CLASS_MODIFICATION_DICT[
                accidental
            ]
        except KeyError:
            message = (
                "Can't initialise WesternPitch with unknown accidental {}!".format(
                    accidental
                )
            )
            raise NotImplementedError(message)

    @staticmethod
    def _translate_pitch_class_name_to_pitch_class(
        pitch_class_name: str,
    ) -> float:
        """Helper function to translate a pitch class name to its respective number.

        +/-1 is defined as one chromatic step. Smaller floating point numbers
        represent microtonal inflections..
        """
        diatonic_pitch_class_name, accidental = (
            pitch_class_name[0],
            pitch_class_name[1:],
        )
        diatonic_pitch_class = (
            parameters.pitches_constants.DIATONIC_PITCH_NAME_TO_PITCH_CLASS_DICT[
                diatonic_pitch_class_name
            ]
        )
        pitch_class_modification = (
            WesternPitch._translate_accidental_to_pitch_class_modifications(accidental)
        )
        return float(diatonic_pitch_class + pitch_class_modification)

    @staticmethod
    def _translate_difference_to_closest_diatonic_pitch_to_accidental(
        difference_to_closest_diatonic_pitch: constants.Real,
    ) -> str:
        """Helper function to translate a number to the closest known accidental."""

        closest_pitch_class_modification: fractions.Fraction = tools.find_closest_item(
            difference_to_closest_diatonic_pitch,
            tuple(
                parameters.pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME_DICT.keys()
            ),
        )
        closest_accidental = parameters.pitches_constants.PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME_DICT[
            closest_pitch_class_modification
        ]
        return closest_accidental

    @staticmethod
    def _translate_pitch_class_to_pitch_class_name(
        pitch_class: constants.Real,
        previous_pitch_class: typing.Optional[constants.Real] = None,
        previous_pitch_class_name: typing.Optional[str] = None,
    ) -> str:
        """Helper function to translate a pitch class in number to a string.

        The returned pitch class name uses a Western nomenclature of English
        diatonic note names. Accidental names are defined in
        mutwo.parameters.pitches_constants.ACCIDENTAL_NAME_TO_PITCH_CLASS_MODIFICATION_DICT.
        For floating point numbers the closest accidental will be chosen.
        """

        # NOTE: it is quite difficult to estimate a new pitch class name from the previous pitch
        # class name. This doesn't work perfect. If this should work someday, you will need
        # to write something like a "Interval" class.

        if previous_pitch_class_name and previous_pitch_class and pitch_class != 0:
            previous_diatonic_pitch_class_name = previous_pitch_class_name[0]
            previous_diatonic_pitch_class_index = (
                parameters.pitches_constants.ASCENDING_DIATONIC_PITCH_NAME_TUPLE.index(
                    previous_diatonic_pitch_class_name
                )
            )
            previous_diatonic_pitch_class = (
                parameters.pitches_constants.DIATONIC_PITCH_NAME_TO_PITCH_CLASS_DICT[
                    previous_diatonic_pitch_class_name
                ]
            )
            n_pitch_classes_difference = pitch_class - previous_pitch_class
            n_diatonic_pitches_to_move = bisect.bisect_left(
                tuple(
                    parameters.pitches_constants.DIATONIC_PITCH_NAME_TO_PITCH_CLASS_DICT.values()
                ),
                n_pitch_classes_difference % 12,
            )
            absolute_new_diatonic_pitch_class_index = (
                previous_diatonic_pitch_class_index + n_diatonic_pitches_to_move
            )
            new_diatonic_pitch_class_index = (
                absolute_new_diatonic_pitch_class_index
                % len(parameters.pitches_constants.ASCENDING_DIATONIC_PITCH_NAME_TUPLE)
            )
            diatonic_pitch = (
                parameters.pitches_constants.ASCENDING_DIATONIC_PITCH_NAME_TUPLE[
                    new_diatonic_pitch_class_index
                ]
            )
            diatonic_pitch_class = (
                parameters.pitches_constants.DIATONIC_PITCH_NAME_TO_PITCH_CLASS_DICT[
                    diatonic_pitch
                ]
            )

            n_pitch_classes_moved_between_diatonic_pitches = (
                diatonic_pitch_class - previous_diatonic_pitch_class
            )
            accidental_adjustments = (
                n_pitch_classes_difference
                - n_pitch_classes_moved_between_diatonic_pitches
            )

            previous_accidental_adjustments = (
                previous_pitch_class - previous_diatonic_pitch_class
            )
            accidental_adjustments += previous_accidental_adjustments

            n_octaves_difference = absolute_new_diatonic_pitch_class_index // len(
                parameters.pitches_constants.ASCENDING_DIATONIC_PITCH_NAME_TUPLE
            )

        else:
            diatonic_pitch_classes = tuple(
                parameters.pitches_constants.DIATONIC_PITCH_NAME_TO_PITCH_CLASS_DICT.values()
            )
            closest_diatonic_pitch_class_index = tools.find_closest_index(
                pitch_class, diatonic_pitch_classes
            )
            diatonic_pitch_class = diatonic_pitch_classes[
                closest_diatonic_pitch_class_index
            ]
            diatonic_pitch = tuple(
                parameters.pitches_constants.DIATONIC_PITCH_NAME_TO_PITCH_CLASS_DICT.keys()
            )[closest_diatonic_pitch_class_index]

            accidental_adjustments = pitch_class - diatonic_pitch_class

        accidental = (
            WesternPitch._translate_difference_to_closest_diatonic_pitch_to_accidental(
                accidental_adjustments
            )
        )

        pitch_class_name = f"{diatonic_pitch}{accidental}"
        return pitch_class_name

    @classmethod
    def from_midi_pitch_number(cls, midi_pitch_number: float) -> WesternPitch:
        pitch_number = midi_pitch_number - 12
        pitch_class_number = pitch_number % 12
        octave_number = pitch_number // 12
        return cls(pitch_class_number, octave=octave_number)

    def __repr__(self) -> str:
        return "{}({})".format(type(self).__name__, self.name)

    def _math(
        self,
        n_pitch_classes_difference: constants.Real,
        operator: typing.Callable[[constants.Real, constants.Real], constants.Real],
    ) -> None:
        new_pitch_class = operator(self.pitch_class, n_pitch_classes_difference)
        n_octaves_difference = new_pitch_class // self.n_pitch_classes_per_octave
        new_pitch_class = new_pitch_class % self.n_pitch_classes_per_octave
        new_octave = self.octave + n_octaves_difference
        self.pitch_class = new_pitch_class
        self.octave = int(new_octave)

    @property
    def name(self) -> str:
        """The name of the pitch in Western nomenclature."""

        return "{}{}".format(self._pitch_class_name, self.octave)

    @property
    def pitch_class_name(self) -> str:
        """The name of the pitch class in Western nomenclature.

        Mutwo uses the English nomenclature for pitch class names:
            (c, d, e, f, g, a, b)
        """

        return self._pitch_class_name

    @pitch_class_name.setter
    def pitch_class_name(self, pitch_class_name: str):
        self._pitch_class = self._translate_pitch_class_name_to_pitch_class(
            pitch_class_name
        )
        self._pitch_class_name = pitch_class_name

    @EqualDividedOctavePitch.pitch_class.setter  # type: ignore
    def pitch_class(self, pitch_class: constants.Real):
        if hasattr(self, "_pitch_class_name"):
            previous_pitch_class = self._pitch_class
            previous_pitch_class_name = self._pitch_class_name
        else:
            previous_pitch_class = None
            previous_pitch_class_name = None

        self._pitch_class_name = self._translate_pitch_class_to_pitch_class_name(
            pitch_class, previous_pitch_class, previous_pitch_class_name
        )
        self._pitch_class = pitch_class



"""Calculate loudness from amplitude."""

import math

import expenvelope  # type: ignore

import mutwo_third_party

from mutwo.core import converters
from mutwo.core.converters.symmetrical import loudness_constants
from mutwo.core.utilities import constants


__all__ = ("LoudnessToAmplitudeConverter",)


class LoudnessToAmplitudeConverter(converters.abc.Converter):
    """Make an approximation of the needed amplitude for a perceived Loudness.

    :param loudspeaker_frequency_response: Optionally the frequency response
        of the used loudspeaker can be added for balancing out uneven curves in
        the loudspeakers frequency response. The frequency response is defined
        with an ``expenvelope.Envelope`` object.
    :type loudspeaker_frequency_response: expenvelope.Envelope
    :param interpolation_order: The interpolation order of the equal loudness
        contour interpolation.
    :type interpolation_order: int

    The converter works best with pure sine waves.
    """

    def __init__(
        self,
        loudspeaker_frequency_response: expenvelope.Envelope = expenvelope.Envelope.from_points(
            (0, 80), (2000, 80)
        ),
        interpolation_order: int = 4,
    ):
        self._interpolation_order = interpolation_order
        self._loudspeaker_frequency_response = loudspeaker_frequency_response
        self._loudspeaker_frequency_response_average = (
            loudspeaker_frequency_response.average_level()
        )

    # ###################################################################### #
    #                          static methods                                #
    # ###################################################################### #

    @staticmethod
    def _decibel_to_amplitude_ratio(
        decibel: constants.Real, reference_amplitude: constants.Real = 1
    ) -> float:
        return float(reference_amplitude * (10 ** (decibel / 20)))

    @staticmethod
    def _decibel_to_power_ratio(decibel: constants.Real) -> float:
        return float(10 ** (decibel / 10))

    @staticmethod
    def _sone_to_phon(loudness_in_sone: constants.Real) -> constants.Real:
        # formula from http://www.sengpielaudio.com/calculatorSonephon.htm
        if loudness_in_sone >= 1:
            return 40 + (10 * math.log(loudness_in_sone, 2))
        else:
            return 40 * (loudness_in_sone + 0.0005) ** 0.35

    # ###################################################################### #
    #               public methods for interaction with the user             #
    # ###################################################################### #

    def convert(
        self, perceived_loudness_in_sone: constants.Real, frequency: constants.Real
    ) -> constants.Real:
        """Calculates the needed amplitude to reach a particular loudness for the entered frequency.

        :param perceived_loudness_in_sone: The subjectively perceived loudness that
            the resulting signal shall have (in the unit `Sone`).
        :type perceived_loudness_in_sone: constants.Real
        :param frequency: A frequency in Hertz for which the necessary amplitude
            shall be calculated.
        :return: Return the amplitude for a sine tone to reach the converters
            loudness when played with the entered frequency.

        **Example:**

        >>> from mutwo.converters import symmetrical
        >>> loudness_converter = symmetrical.loudness.LoudnessToAmplitudeConverter(1)
        >>> loudness_converter.convert(200)
        0.009364120303317933
        >>> loudness_converter.convert(50)
        0.15497924558613232
        """

        perceived_loudness_in_phon = self._sone_to_phon(perceived_loudness_in_sone)
        equal_loudness_contour_interpolation = mutwo_third_party.pydsm.pydsm.iso226.iso226_spl_itpl(  # type: ignore
            perceived_loudness_in_phon, self._interpolation_order
        )

        # (1) calculates necessary sound pressure level depending on the frequency
        #     and loudness (to get the same loudness over all frequencies)
        sound_pressure_level_for_perceived_loudness_based_on_frequency = float(
            equal_loudness_contour_interpolation(frequency)
        )
        # (2) figure out the produced soundpressure of the loudspeaker depending
        #     on the frequency (for balancing uneven frequency responses of
        #     loudspeakers)
        produced_soundpressure_for_1_watt_1_meter_depending_on_loudspeaker = (
            self._loudspeaker_frequency_response.value_at(frequency)
        )
        difference_to_average = (
            self._loudspeaker_frequency_response_average
            - produced_soundpressure_for_1_watt_1_meter_depending_on_loudspeaker
        )
        sound_pressure_level_for_pereived_loudness_based_on_speaker = (
            sound_pressure_level_for_perceived_loudness_based_on_frequency
            + difference_to_average
        )
        amplitude_ratio = self._decibel_to_amplitude_ratio(
            sound_pressure_level_for_pereived_loudness_based_on_speaker,
            loudness_constants.AUDITORY_THRESHOLD_AT_1KHZ,
        )
        return amplitude_ratio



"""Abstract base classes for different parameters.

This module defines the public API of parameters.
Most other mutwo classes rely on this API. This means
when someone creates a new class inheriting from any of the
abstract parameter classes which are defined in this module,
she or he can make use of all other mutwo modules with this
newly created parameter class.
"""

from __future__ import annotations

import abc
import dataclasses
import functools
import math
import typing

try:
    import quicktions as fractions  # type: ignore
except ImportError:
    import fractions  # type: ignore

from mutwo.core import events
from mutwo.core.parameters import pitches_constants
from mutwo.core.parameters import volumes_constants
from mutwo.core.utilities import constants
from mutwo.core.utilities import decorators
from mutwo.core.utilities import tools

__all__ = (
    "ParameterWithEnvelope",
    "PitchInterval",
    "Pitch",
    "Volume",
    "PlayingIndicator",
    "NotationIndicator",
)

DurationType = constants.Real


class ParameterWithEnvelope(abc.ABC):
    """Abstract base class for all parameters with an envelope."""

    def __init__(self, envelope: events.envelopes.RelativeEnvelope):
        self.envelope = envelope

    @property
    def envelope(self) -> events.envelopes.RelativeEnvelope:
        return self._envelope

    @envelope.setter
    def envelope(self, new_envelope: typing.Any):
        try:
            assert isinstance(new_envelope, events.envelopes.RelativeEnvelope)
        except AssertionError:
            raise TypeError(
                f"Found illegal object '{new_envelope}' of not "
                f"supported type '{type(new_envelope)}'. "
                f"Only instances of '{events.envelopes.RelativeEnvelope}'"
                " are allowed!"
            )
        self._envelope = new_envelope

    def resolve_envelope(
        self,
        duration: constants.DurationType,
        resolve_envelope_class: type[
            events.envelopes.Envelope
        ] = events.envelopes.Envelope,
    ) -> events.envelopes.Envelope:
        return self.envelope.resolve(duration, self, resolve_envelope_class)


@functools.total_ordering  # type: ignore
class PitchInterval(abc.ABC):
    """Abstract base class for any pitch interval class

    If the user wants to define a new pitch class, the abstract
    property :attr:`cents` has to be overridden.

    See `wikipedia entry <https://en.wikipedia.org/wiki/Cent_(music)>`_
    for definition of 'cents'.
    """

    @property
    @abc.abstractmethod
    def cents(self) -> float:
        raise NotImplementedError

    def __eq__(self, other: typing.Any) -> bool:
        try:
            return self.cents == other.cents
        except AttributeError:
            return False

    def __lt__(self, other: PitchInterval) -> bool:
        return self.cents < other.cents


@functools.total_ordering  # type: ignore
class Pitch(ParameterWithEnvelope):
    """Abstract base class for any pitch class.

    If the user wants to define a new pitch class, the abstract
    property :attr:`frequency` has to be overridden. Starting
    from mutwo version = 0.46.0 the user will furthermore have
    to define an :func:`add` and a :func:`subtract` method.
    """

    class PitchEnvelope(events.envelopes.Envelope):
        """Default resolution envelope class for :class:`Pitch`"""

        def __init__(
            self,
            *args,
            event_to_parameter: typing.Optional[
                typing.Callable[[events.abc.Event], constants.ParameterType]
            ] = None,
            value_to_parameter: typing.Optional[
                typing.Callable[
                    [events.envelopes.Envelope.Value], constants.ParameterType
                ]
            ] = None,
            parameter_to_value: typing.Optional[
                typing.Callable[
                    [constants.ParameterType], events.envelopes.Envelope.Value
                ]
            ] = None,
            apply_parameter_on_event: typing.Optional[
                typing.Callable[[events.abc.Event, constants.ParameterType], None]
            ] = None,
            **kwargs,
        ):
            if not event_to_parameter:
                event_to_parameter = self._event_to_parameter
            if not value_to_parameter:
                value_to_parameter = self._value_to_parameter
            if not apply_parameter_on_event:
                apply_parameter_on_event = self._apply_parameter_on_event
            if not parameter_to_value:
                parameter_to_value = self._parameter_to_value

            super().__init__(
                *args,
                event_to_parameter=event_to_parameter,
                value_to_parameter=value_to_parameter,
                parameter_to_value=parameter_to_value,
                apply_parameter_on_event=apply_parameter_on_event,
                **kwargs,
            )

        @classmethod
        def make_generic_pitch_class(cls, frequency: constants.Real) -> type[Pitch]:
            @decorators.add_copy_option
            def generic_add(self, pitch_interval: PitchInterval) -> Pitch:
                self.frequency = (
                    Pitch.cents_to_ratio(pitch_interval.cents) * self.frequency
                )
                return self

            @decorators.add_copy_option
            def generic_subtract(self, pitch_interval: PitchInterval) -> Pitch:
                self.frequency = self.frequency / Pitch.cents_to_ratio(
                    pitch_interval.cents
                )
                return self

            generic_pitch_class = type(
                "GenericPitch",
                (Pitch,),
                {
                    "frequency": frequency,
                    "add": generic_add,
                    "subtract": generic_subtract,
                    "__repr__": lambda self: f"GenericPitchInterval(frequency = {self.frequency})",
                },
            )

            return generic_pitch_class

        @classmethod
        def make_generic_pitch(cls, frequency: constants.Real) -> Pitch:
            return cls.make_generic_pitch_class(frequency)

        @classmethod
        def _value_to_parameter(
            cls,
            value: events.envelopes.Envelope.Value,  # type: ignore
        ) -> constants.ParameterType:
            # For inner calculation (value) cents are used instead
            # of frequencies. In this way we can ensure that the transitions
            # are closer to the human logarithmic hearing.
            # See als `_parameter_to_value`.
            frequency = (
                Pitch.cents_to_ratio(value)
                * pitches_constants.PITCH_ENVELOPE_REFERENCE_FREQUENCY
            )
            return cls.make_generic_pitch(frequency)

        @classmethod
        def _event_to_parameter(
            cls, event: events.abc.Event
        ) -> constants.ParameterType:
            if hasattr(
                event,
                pitches_constants.DEFAULT_PITCH_ENVELOPE_PARAMETER_NAME,
            ):
                return getattr(
                    event,
                    pitches_constants.DEFAULT_PITCH_ENVELOPE_PARAMETER_NAME,
                )
            else:
                return cls.make_generic_pitch(pitches_constants.DEFAULT_CONCERT_PITCH)

        @classmethod
        def _apply_parameter_on_event(
            cls, event: events.abc.Event, parameter: constants.ParameterType
        ):
            setattr(
                event,
                pitches_constants.DEFAULT_PITCH_ENVELOPE_PARAMETER_NAME,
                parameter,
            )

        @classmethod
        def _parameter_to_value(
            cls, parameter: constants.ParameterType
        ) -> constants.Real:
            # For inner calculation (value) cents are used instead
            # of frequencies. In this way we can ensure that the transitions
            # are closer to the human logarithmic hearing.
            # See als `_value_to_parameter`.
            return Pitch.hertz_to_cents(
                pitches_constants.PITCH_ENVELOPE_REFERENCE_FREQUENCY,
                parameter.frequency,
            )

    class PitchIntervalEnvelope(events.envelopes.RelativeEnvelope):
        """Default envelope class for :class:`Pitch`

        Resolves into :class:`Pitch.PitchEnvelope`.
        """

        def __init__(
            self,
            *args,
            event_to_parameter: typing.Optional[
                typing.Callable[[events.abc.Event], constants.ParameterType]
            ] = None,
            value_to_parameter: typing.Optional[
                typing.Callable[
                    [events.envelopes.Envelope.Value], constants.ParameterType
                ]
            ] = None,
            parameter_to_value: typing.Callable[
                [constants.ParameterType], events.envelopes.Envelope.Value
            ] = lambda parameter: parameter.cents,
            apply_parameter_on_event: typing.Optional[
                typing.Callable[[events.abc.Event, constants.ParameterType], None]
            ] = None,
            base_parameter_and_relative_parameter_to_absolute_parameter: typing.Optional[
                typing.Callable[
                    [constants.ParameterType, constants.ParameterType],
                    constants.ParameterType,
                ]
            ] = None,
            **kwargs,
        ):
            if not event_to_parameter:
                event_to_parameter = self._event_to_parameter
            if not value_to_parameter:
                value_to_parameter = self._value_to_parameter
            if not apply_parameter_on_event:
                apply_parameter_on_event = self._apply_parameter_on_event
            if not base_parameter_and_relative_parameter_to_absolute_parameter:
                base_parameter_and_relative_parameter_to_absolute_parameter = (
                    self._base_parameter_and_relative_parameter_to_absolute_parameter
                )

            super().__init__(
                *args,
                event_to_parameter=event_to_parameter,
                value_to_parameter=value_to_parameter,
                parameter_to_value=parameter_to_value,
                apply_parameter_on_event=apply_parameter_on_event,
                base_parameter_and_relative_parameter_to_absolute_parameter=base_parameter_and_relative_parameter_to_absolute_parameter,
                **kwargs,
            )

        @staticmethod
        def make_generic_pitch_interval(cents: constants.Real) -> PitchInterval:
            return type(
                "GenericPitchInterval",
                (PitchInterval,),
                {
                    "cents": cents,
                    "__repr__": lambda self: f"GenericPitchInterval(cents = {self.cents})",
                },
            )()

        @classmethod
        def _event_to_parameter(
            cls, event: events.abc.Event
        ) -> constants.ParameterType:
            if hasattr(
                event, pitches_constants.DEFAULT_PITCH_INTERVAL_ENVELOPE_PARAMETER_NAME
            ):
                return getattr(
                    event,
                    pitches_constants.DEFAULT_PITCH_INTERVAL_ENVELOPE_PARAMETER_NAME,
                )
            else:
                return cls.make_generic_pitch_interval(0)

        @classmethod
        def _value_to_parameter(
            cls, value: events.envelopes.Envelope.Value
        ) -> constants.ParameterType:
            return cls.make_generic_pitch_interval(value)

        @classmethod
        def _apply_parameter_on_event(
            cls, event: events.abc.Event, parameter: constants.ParameterType
        ):
            setattr(
                event,
                pitches_constants.DEFAULT_PITCH_INTERVAL_ENVELOPE_PARAMETER_NAME,
                parameter,
            ),

        @classmethod
        def _base_parameter_and_relative_parameter_to_absolute_parameter(
            cls, base_parameter: Pitch, relative_parameter: PitchInterval
        ) -> Pitch:
            return base_parameter + relative_parameter

    def __init__(self, envelope: typing.Optional[Pitch.PitchIntervalEnvelope] = None):
        if not envelope:
            generic_pitch_interval = (
                self.PitchIntervalEnvelope.make_generic_pitch_interval(0)
            )
            envelope = self.PitchIntervalEnvelope([[0, generic_pitch_interval]])
        super().__init__(envelope)

    # ###################################################################### #
    #     conversion methods between different pitch describing units        #
    # ###################################################################### #

    @staticmethod
    def hertz_to_cents(frequency0: constants.Real, frequency1: constants.Real) -> float:
        """Calculates the difference in cents between two frequencies.

        :param frequency0: The first frequency in Hertz.
        :param frequency1: The second frequency in Hertz.
        :return: The difference in cents between the first and the second
            frequency.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Pitch.hertz_to_cents(200, 400)
        1200.0
        """
        return float(1200 * math.log(frequency1 / frequency0, 2))

    @staticmethod
    def ratio_to_cents(ratio: fractions.Fraction) -> float:
        """Converts a frequency ratio to its respective cent value.

        :param ratio: The frequency ratio which cent value shall be
            calculated.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Pitch.ratio_to_cents(fractions.Fraction(3, 2))
        701.9550008653874
        """
        return pitches_constants.CENT_CALCULATION_CONSTANT * math.log10(ratio)

    @staticmethod
    def cents_to_ratio(cents: constants.Real) -> fractions.Fraction:
        """Converts a cent value to its respective frequency ratio.

        :param cents: Cents that shall be converted to a frequency ratio.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Pitch.cents_to_ratio(1200)
        Fraction(2, 1)
        """
        return fractions.Fraction(
            10 ** (cents / pitches_constants.CENT_CALCULATION_CONSTANT)
        )

    @staticmethod
    def hertz_to_midi_pitch_number(frequency: constants.Real) -> float:
        """Converts a frequency in hertz to its respective midi pitch.

        :param frequency: The frequency that shall be translated to a midi pitch
            number.
        :return: The midi pitch number (potentially a floating point number if the
            entered frequency isn't on the grid of the equal divided octave tuning
            with a = 440 Hertz).

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Pitch.hertz_to_midi_pitch_number(440)
        69.0
        >>> abc.Pitch.hertz_to_midi_pitch_number(440 * 3 / 2)
        75.98044999134612
        """
        closest_frequency_index = tools.find_closest_index(
            frequency, pitches_constants.MIDI_PITCH_FREQUENCY_TUPLE
        )
        closest_frequency = pitches_constants.MIDI_PITCH_FREQUENCY_TUPLE[
            closest_frequency_index
        ]
        closest_midi_pitch_number = pitches_constants.MIDI_PITCH_NUMBER_TUPLE[
            closest_frequency_index
        ]
        difference_in_cents = Pitch.hertz_to_cents(frequency, closest_frequency)
        return float(closest_midi_pitch_number + (difference_in_cents / 100))

    # ###################################################################### #
    #                            public properties                           #
    # ###################################################################### #

    @property
    @abc.abstractmethod
    def frequency(self) -> float:
        """The frequency in Hertz of the pitch."""
        raise NotImplementedError

    @property
    def midi_pitch_number(self) -> float:
        """The midi pitch number (from 0 to 127) of the pitch."""
        return self.hertz_to_midi_pitch_number(self.frequency)

    # ###################################################################### #
    #                            comparison methods                          #
    # ###################################################################### #

    def __lt__(self, other: Pitch) -> bool:
        return self.frequency < other.frequency

    def __eq__(self, other: object) -> bool:
        try:
            return self.frequency == other.frequency  # type: ignore
        except AttributeError:
            return False

    @abc.abstractmethod
    def add(self, pitch_interval: PitchInterval, mutate: bool = True) -> Pitch:
        raise NotImplementedError

    @abc.abstractmethod
    def subtract(self, pitch_interval: PitchInterval, mutate: bool = True) -> Pitch:
        raise NotImplementedError

    def __add__(self, pitch_interval: PitchInterval) -> Pitch:
        return self.add(pitch_interval, mutate=False)

    def __sub__(self, pitch_interval: PitchInterval) -> Pitch:
        return self.subtract(pitch_interval, mutate=False)

    def resolve_envelope(
        self,
        duration: constants.DurationType,
        resolve_envelope_class: typing.Optional[type[events.envelopes.Envelope]] = None,
    ) -> events.envelopes.Envelope:
        if not resolve_envelope_class:
            resolve_envelope_class = Pitch.PitchEnvelope
        return super().resolve_envelope(duration, resolve_envelope_class)


@functools.total_ordering  # type: ignore
class Volume(abc.ABC):
    """Abstract base class for any volume class.

    If the user wants to define a new volume class, the abstract
    property :attr:`` has to be overridden.
    """

    @staticmethod
    def decibel_to_amplitude_ratio(
        decibel: constants.Real, reference_amplitude: constants.Real = 1
    ) -> float:
        """Convert decibel to amplitude ratio.

        :param decibel: The decibel number that shall be converted.
        :param reference_amplitude: The amplitude for decibel == 0.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Volume.decibel_to_amplitude_ratio(0)
        1
        >>> abc.Volume.decibel_to_amplitude_ratio(-6)
        0.5011872336272722
        >>> abc.Volume.decibel_to_amplitude_ratio(0, reference_amplitude=0.25)
        0.25
        """
        return float(reference_amplitude * (10 ** (decibel / 20)))

    @staticmethod
    def decibel_to_power_ratio(
        decibel: constants.Real, reference_amplitude: constants.Real = 1
    ) -> float:
        """Convert decibel to power ratio.

        :param decibel: The decibel number that shall be converted.
        :param reference_amplitude: The amplitude for decibel == 0.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Volume.decibel_to_power_ratio(0)
        1
        >>> abc.Volume.decibel_to_power_ratio(-6)
        0.251188643150958
        >>> abc.Volume.decibel_to_power_ratio(0, reference_amplitude=0.25)
        0.25
        """
        return float(reference_amplitude * (10 ** (decibel / 10)))

    @staticmethod
    def amplitude_ratio_to_decibel(
        amplitude: constants.Real, reference_amplitude: constants.Real = 1
    ) -> float:
        """Convert amplitude ratio to decibel.

        :param amplitude: The amplitude that shall be converted.
        :param reference_amplitude: The amplitude for decibel == 0.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Volume.amplitude_ratio_to_decibel(1)
        0
        >>> abc.Volume.amplitude_ratio_to_decibel(0)
        inf
        >>> abc.Volume.amplitude_ratio_to_decibel(0.5)
        -6.020599913279624
        """
        if amplitude == 0:
            return float("-inf")
        else:
            return float(20 * math.log10(amplitude / reference_amplitude))

    @staticmethod
    def power_ratio_to_decibel(
        amplitude: constants.Real, reference_amplitude: constants.Real = 1
    ) -> float:
        """Convert power ratio to decibel.

        :param amplitude: The amplitude that shall be converted.
        :param reference_amplitude: The amplitude for decibel == 0.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Volume.power_ratio_to_decibel(1)
        0
        >>> abc.Volume.power_ratio_to_decibel(0)
        inf
        >>> abc.Volume.power_ratio_to_decibel(0.5)
        -3.010299956639812
        """
        if amplitude == 0:
            return float("-inf")
        else:
            return float(10 * math.log10(amplitude / reference_amplitude))

    @staticmethod
    def amplitude_ratio_to_midi_velocity(
        amplitude: constants.Real, reference_amplitude: constants.Real = 1
    ) -> int:
        """Convert amplitude ratio to midi velocity.

        :param amplitude: The amplitude which shall be converted.
        :type amplitude: constants.Real
        :param reference_amplitude: The amplitude for decibel == 0.
        :return: The midi velocity.

        The method clips values that are higher than 1 / lower than 0.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Volume.amplitude_ratio_to_midi_velocity(1)
        127
        >>> abc.Volume.amplitude_ratio_to_midi_velocity(0)
        0
        """

        return Volume.decibel_to_midi_velocity(
            Volume.amplitude_ratio_to_decibel(
                amplitude, reference_amplitude=reference_amplitude
            )
        )

    @staticmethod
    def decibel_to_midi_velocity(
        decibel_to_convert: constants.Real,
        minimum_decibel: typing.Optional[constants.Real] = None,
        maximum_decibel: typing.Optional[constants.Real] = None,
    ) -> int:
        """Convert decibel to midi velocity (0 to 127).

        :param decibel: The decibel value which shall be converted..
        :type decibel: constants.Real
        :param minimum_decibel: The decibel value which is equal to the lowest
            midi velocity (0).
        :type minimum_decibel: constants.Real, optional
        :param maximum_decibel: The decibel value which is equal to the highest
            midi velocity (127).
        :type maximum_decibel: constants.Real, optional
        :return: The midi velocity.

        The method clips values which are higher than 'maximum_decibel' and lower than
        'minimum_decibel'.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Volume.decibel_to_midi_velocity(0)
        127
        >>> abc.Volume.decibel_to_midi_velocity(-40)
        0
        """

        if minimum_decibel is None:
            minimum_decibel = (
                volumes_constants.DEFAULT_MINIMUM_DECIBEL_FOR_MIDI_VELOCITY_AND_STANDARD_DYNAMIC_INDICATOR
            )

        if maximum_decibel is None:
            maximum_decibel = (
                volumes_constants.DEFAULT_MAXIMUM_DECIBEL_FOR_MIDI_VELOCITY_AND_STANDARD_DYNAMIC_INDICATOR
            )

        if decibel_to_convert > maximum_decibel:
            decibel_to_convert = maximum_decibel

        if decibel_to_convert < minimum_decibel:
            decibel_to_convert = minimum_decibel

        velocity = int(
            tools.scale(
                decibel_to_convert,
                minimum_decibel,
                maximum_decibel,
                volumes_constants.MINIMUM_VELOCITY,
                volumes_constants.MAXIMUM_VELOCITY,
            )
        )

        return velocity

    # properties
    @property
    @abc.abstractmethod
    def amplitude(self) -> constants.Real:
        """The amplitude of the Volume (a number from 0 to 1)."""
        raise NotImplementedError

    @property
    def decibel(self) -> constants.Real:
        """The decibel of the volume (from -120 to 0)"""
        return self.amplitude_ratio_to_decibel(self.amplitude)

    @property
    def midi_velocity(self) -> int:
        """The velocity of the volume (from 0 to 127)."""
        return self.decibel_to_midi_velocity(self.decibel)

    # comparison methods
    def __lt__(self, other: Volume) -> bool:
        return self.amplitude < other.amplitude

    def __eq__(self, other: object) -> bool:
        try:
            return self.amplitude == other.amplitude  # type: ignore
        except AttributeError:
            return False


@dataclasses.dataclass()  # type: ignore
class Indicator(abc.ABC):
    @property
    @abc.abstractmethod
    def is_active(self) -> bool:
        raise NotImplementedError()

    def get_arguments_dict(self) -> dict[str, typing.Any]:
        return {
            key: getattr(self, key)
            for key in self.__dataclass_fields__.keys()  # type: ignore
        }


class PlayingIndicator(Indicator):
    """Abstract base class for any playing indicator."""

    pass


class ExplicitPlayingIndicator(PlayingIndicator):
    def __init__(self, is_active: bool = False):
        self.is_active = is_active

    def __repr__(self):
        return "{}({})".format(type(self).__name__, self.is_active)

    def get_arguments_dict(self) -> dict[str, typing.Any]:
        return {"is_active": self.is_active}

    @property
    def is_active(self) -> bool:
        return self._is_active

    @is_active.setter
    def is_active(self, is_active: bool):
        self._is_active = is_active


@dataclasses.dataclass()
class ImplicitPlayingIndicator(PlayingIndicator):
    @property
    def is_active(self) -> bool:
        return all(
            tuple(
                argument is not None for argument in self.get_arguments_dict().values()
            )
        )


class NotationIndicator(Indicator):
    """Abstract base class for any notation indicator."""

    @property
    def is_active(self) -> bool:
        return all(
            tuple(
                argument is not None for argument in self.get_arguments_dict().values()
            )
        )


T = typing.TypeVar("T", PlayingIndicator, NotationIndicator)


@dataclasses.dataclass
class IndicatorCollection(typing.Generic[T]):
    def get_all_indicator(self) -> tuple[T, ...]:
        return tuple(
            getattr(self, key)
            for key in self.__dataclass_fields__.keys()  # type: ignore
        )

    def get_indicator_dict(self) -> dict[str, Indicator]:
        return {key: getattr(self, key) for key in self.__dataclass_fields__.keys()}  # type: ignore



"""Abstractions for attributes that can be assigned to Event objects.

"""

from . import commas

from . import pitches_constants
from . import playing_indicators_constants
from . import volumes_constants

from . import abc

from . import notation_indicators
from . import pitches
from . import tempos
from . import playing_indicators
from . import volumes



"""Convert mutwo or generic Python objects to other mutwo or generic Python objects."""

from . import loudness_constants

from . import grace_notes
from . import loudness
from . import metricities
from . import playing_indicators
from . import spectrals
from . import tempos



"""Submodule for the parameter volume.

'Volume' is defined as any object that knows a :attr:`amplitude` attribute.
"""

import typing

import numpy as np  # type: ignore

from mutwo.core import parameters
from mutwo.core.utilities import constants
from mutwo.core.utilities import tools

__all__ = ("DirectVolume", "DecibelVolume", "WesternVolume")


class DirectVolume(parameters.abc.Volume):
    """A simple volume class that gets directly initialised by its amplitude.

    :param amplitude: The amplitude of the :class:`DirectVolume` object.

    May be used when a converter class needs a volume object, but there is
    no need or desire for a complex abstraction of the respective volume.
    """

    def __init__(self, amplitude: constants.Real):
        self._amplitude = amplitude

    @property
    def amplitude(self) -> constants.Real:
        return self._amplitude

    def __repr__(self) -> str:
        return "{}({})".format(type(self).__name__, self.amplitude)


class DecibelVolume(parameters.abc.Volume):
    """A simple volume class that gets directly initialised by decibel.

    :param decibel: The decibel of the :class:`DecibelVolume` object (should be
        from -120 to 0).

    May be used when a converter class needs a volume object, but there is
    no need or desire for a complex abstraction of the respective volume.
    """

    def __init__(self, decibel: constants.Real):
        self._decibel = decibel

    @property
    def decibel(self) -> constants.Real:
        return self._decibel

    @property
    def amplitude(self) -> constants.Real:
        return self.decibel_to_amplitude_ratio(self.decibel)

    def __repr__(self) -> str:
        return "{}({})".format(type(self).__name__, self.amplitude)


class WesternVolume(parameters.abc.Volume):
    """Volume with a traditional Western nomenclature.

    :param name: Dynamic indicator in traditional Western nomenclature
        ('f', 'pp', 'mf', 'sfz', etc.). For a list of all supported
        indicators, see :const:`mutwo.parameters.volumes_constants.DYNAMIC_INDICATOR_TUPLE`.
    :type name: str
    :param minimum_decibel: The decibel value which is equal to the lowest dynamic indicator
        (ppppp).
    :type minimum_decibel: constants.Real, optional
    :param maximum_decibel: The decibel value which is equal to the highest dynamic indicator
        (fffff).
    :type maximum_decibel: constants.Real, optional

    **Example:**

    >>> from mutwo.parameters import volumes
    >>> volumes.WesternVolume('fff')
    WesternVolume(fff)
    """

    def __init__(
        self,
        name: str,
        minimum_decibel: typing.Optional[constants.Real] = None,
        maximum_decibel: typing.Optional[constants.Real] = None,
    ):
        if minimum_decibel is None:
            minimum_decibel = (
                parameters.volumes_constants.DEFAULT_MINIMUM_DECIBEL_FOR_MIDI_VELOCITY_AND_STANDARD_DYNAMIC_INDICATOR
            )

        if maximum_decibel is None:
            maximum_decibel = (
                parameters.volumes_constants.DEFAULT_MAXIMUM_DECIBEL_FOR_MIDI_VELOCITY_AND_STANDARD_DYNAMIC_INDICATOR
            )

        self.name = name
        self._standard_dynamic_indicator_to_decibel_mapping = WesternVolume._make_standard_dynamic_indicator_to_value_dict(
            minimum_decibel, maximum_decibel, float,
        )
        self._dynamic_indicator_to_decibel_mapping = WesternVolume._make_dynamic_indicator_to_value_dict(
            self._standard_dynamic_indicator_to_decibel_mapping
        )
        self._decibel_to_standard_dynamic_indicator_mapping = {
            decibel: dynamic_indicator
            for dynamic_indicator, decibel in self._standard_dynamic_indicator_to_decibel_mapping.items()
        }

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.name})"

    # ###################################################################### #
    #                      static private methods                            #
    # ###################################################################### #

    @staticmethod
    def _make_standard_dynamic_indicator_to_value_dict(
        minima: float, maxima: float, dtype: typing.Type[float] = float
    ) -> dict[str, float]:
        return {
            dynamic_indicator: decibel
            for dynamic_indicator, decibel in zip(
                parameters.volumes_constants.STANDARD_DYNAMIC_INDICATOR,
                np.linspace(
                    minima,
                    maxima,
                    len(parameters.volumes_constants.STANDARD_DYNAMIC_INDICATOR),
                    dtype=dtype,
                ),
            )
        }

    @staticmethod
    def _make_dynamic_indicator_to_value_dict(
        standard_dynamic_indicator_to_value_dict: dict[str, float]
    ) -> dict[str, float]:
        dynamic_indicator_to_value_dict = {}
        dynamic_indicator_to_value_dict.update(
            standard_dynamic_indicator_to_value_dict
        )
        for (
            special_dynamic_indicator,
            standard_dynamic_indicator,
        ) in (
            parameters.volumes_constants.SPECIAL_DYNAMIC_INDICATOR_TO_STANDARD_DYNAMIC_INDICATOR_DICT.items()
        ):
            dynamic_indicator_to_value_dict.update(
                {
                    special_dynamic_indicator: dynamic_indicator_to_value_dict[
                        standard_dynamic_indicator
                    ]
                }
            )
        return dynamic_indicator_to_value_dict

    # ###################################################################### #
    #                class methods (alternative constructors)                #
    # ###################################################################### #

    @classmethod
    def from_amplitude(cls, amplitude: constants.Real) -> "WesternVolume":
        """Initialise `WesternVolume` from amplitude ratio.

        :param amplitude: The amplitude which shall be converted to a `WesternVolume`
            object.

        >>> from mutwo.parameters import volumes
        >>> volumes.WesternVolume.from_amplitude(0.05)
        WesternVolume(mp)
        """
        decibel = cls.amplitude_ratio_to_decibel(amplitude)
        return cls.from_decibel(decibel)

    @classmethod
    def from_decibel(cls, decibel: constants.Real) -> "WesternVolume":
        """Initialise `WesternVolume` from decibel.

        :param decibel: The decibel which shall be converted to a `WesternVolume`
            object.

        >>> from mutwo.parameters import volumes
        >>> volumes.WesternVolume.from_decibel(-24)
        WesternVolume(mf)
        """
        volume_object = cls("mf")
        closest_decibel: float = tools.find_closest_item(
            decibel,
            tuple(volume_object._decibel_to_standard_dynamic_indicator_mapping.keys()),
        )
        indicator = volume_object._decibel_to_standard_dynamic_indicator_mapping[
            closest_decibel
        ]
        volume_object.name = indicator
        return volume_object

    # ###################################################################### #
    #                             properties                                 #
    # ###################################################################### #

    @property
    def name(self) -> str:
        """The western nomenclature name for dynamic.

        For a list of all supported indicators, see
        :const:`mutwo.parameters.volumes_constants.DYNAMIC_INDICATOR_TUPLE`.
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        try:
            assert name in parameters.volumes_constants.DYNAMIC_INDICATOR_TUPLE
        except AssertionError:
            message = (
                "unknown dynamic name '{}'. Supported dynamic names are '{}'.".format(
                    name, parameters.volumes_constants.DYNAMIC_INDICATOR_TUPLE,
                )
            )
            raise ValueError(message)
        self._name = name

    @property
    def decibel(self) -> constants.Real:
        return self._dynamic_indicator_to_decibel_mapping[self.name]

    @property
    def amplitude(self) -> constants.Real:
        return self.decibel_to_amplitude_ratio(self.decibel)



import typing

from mutwo.core import converters
from mutwo.core import parameters


class TwoPitchesToCommonHarmonicsConverter(converters.abc.Converter):
    """Find the common harmonics between two pitches.

    :param tonality: ``True`` for finding common harmonics, ``False`` for finding
        common subharmonics and ``None`` for finding common pitches between the
        harmonics of the first pitch and the subharmonics of the second pitch.
    :type tonality: typing.Optional[bool]
    :param lowest_partial: The lowest partial to get investigated. Shouldn't be smaller
        than 1.
    :type lowest_partial: int
    :param highest_partial: The highest partial to get investigated. Shouldn't be bigger
        than 1.
    :type highest_partial: int
    """

    def __init__(
        self, tonality: typing.Optional[bool], lowest_partial: int, highest_partial: int
    ):
        self._tonality_per_pitch_tuple = (
            (tonality, tonality) if tonality is not None else (True, False)
        )
        self._tonality_to_partial_tuple_dict = {
            tonality: TwoPitchesToCommonHarmonicsConverter._make_partials(
                lowest_partial, highest_partial, tonality
            )
            for tonality in (True, False)
        }

    @staticmethod
    def _make_partials(
        lowest_partial: int, highest_partial: int, tonality: bool
    ) -> tuple[tuple[int, parameters.pitches.JustIntonationPitch], ...]:
        partial_tuple = tuple(
            (nth_partial, parameters.pitches.JustIntonationPitch(nth_partial, 1))
            for nth_partial in range(lowest_partial, highest_partial)
        )
        if not tonality:
            [partial.inverse() for _, partial in partial_tuple]
        return partial_tuple

    def convert(
        self,
        pitch_pair_to_examine: tuple[
            parameters.pitches.JustIntonationPitch,
            parameters.pitches.JustIntonationPitch,
        ],
    ) -> tuple[parameters.pitches.CommonHarmonic, ...]:
        partials0, partials1 = tuple(
            tuple(
                (nth_partial, partial + pitch)
                for nth_partial, partial in self._tonality_to_partial_tuple_dict[tonality]
            )
            for pitch, tonality in zip(pitch_pair_to_examine, self._tonality_per_pitch_tuple)
        )

        nth_partial_for_partials1, partials1 = zip(*partials1)

        common_harmonic_list = []
        for nth_partial_for_first_pitch, partial in partials0:
            if partial in partials1:
                nth_partial_for_second_pitch = nth_partial_for_partials1[
                    partials1.index(partial)
                ]
                common_harmonic = parameters.pitches.CommonHarmonic(
                    tuple(
                        parameters.pitches.Partial(nth_partial, tonality)
                        for nth_partial, tonality in zip(
                            (nth_partial_for_first_pitch, nth_partial_for_second_pitch),
                            self._tonality_per_pitch_tuple,
                        )
                    ),
                    partial,
                )
                common_harmonic_list.append(common_harmonic)

        return tuple(common_harmonic_list)



"""Constants used in and for :mod:`mutwo.core.events.envelopes`"""

DEFAULT_PARAMETER_ATTRIBUTE_NAME = "value"
"""Default attribute name when fetching the parameter of an event"""

DEFAULT_CURVE_SHAPE_ATTRIBUTE_NAME = "curve_shape"
"""Default attribute name when fetching the curve shape of an event"""



"""Constants to be used in `mutwo.parameters.playing_indicator`"""

import typing

ARTICULATION_LITERAL = typing.Literal[
    # Copy/paste from
    # https://abjad.github.io/_modules/abjad/indicators/Articulation.html
    "accent",
    "marcato",
    "staccatissimo",
    "espressivo",
    "staccato",
    "tenuto",
    "portato",
    "upbow",
    "downbow",
    "flageolet",
    "thumb",
    "lheel",
    "rheel",
    "ltoe",
    "rtoe",
    "open",
    "halfopen",
    "snappizzicato",
    "stopped",
    "turn",
    "reverseturn",
    "trill",
    "prall",
    "mordent",
    "prallprall",
    "prallmordent",
    "upprall",
    "downprall",
    "upmordent",
    "downmordent",
    "pralldown",
    "prallup",
    "lineprall",
    "signumcongruentiae",
    "shortfermata",
    "fermata",
    "longfermata",
    "verylongfermata",
    "segno",
    "coda",
    "varcoda",
    "^",
    "+",
    "-",
    "|",
    ">",
    ".",
    "_",
]
"""A sequence of all allowed articulation.

'allowed' means merely that all defined strings here
should work with the `abjad` module in mutwos frontend module.

Copy/paste from
https://abjad.github.io/_modules/abjad/indicators/Articulation.html"""

CONTACT_POINT_LITERAL = typing.Literal[
    # (Mostly) copied from
    # https://abjad.github.io/_modules/abjad/indicators/StringContactPoint.html#StringContactPoint
    "dietro ponticello",
    "molto sul ponticello",
    "molto sul tasto",
    "ordinario",
    "pizzicato",
    "ponticello",
    "sul ponticello",
    "sul tasto",
    "col legno tratto",
    # abbreviations also work
    "d.p.",
    "m.s.p",
    "m.s.t.",
    "ord.",
    "pizz.",
    "p.",
    "s.p.",
    "s.t.",
    "c.l.t.",
]
"""A sequence of all allowed contact points.

'allowed' means merely that all defined strings here
should work with the `abjad` module in mutwos frontend module.

(Mostly) copied from
https://abjad.github.io/_modules/abjad/indicators/StringContactPoint.html#StringContactPoint"""

PEDAL_TYPE_LITERAL = typing.Literal["sustain", "sostenuto", "corda"]
"""A sequence of all allowed pedal types.

'allowed' means merely that all defined strings here
should work with the `abjad` module in mutwos frontend module.

Pedal types copied from
https://abjad.github.io/_modules/abjad/indicators/StartPianoPedal.html"""

FERMATA_TYPE_LITERAL = typing.Literal[
    "shortfermata",
    "fermata",
    "longfermata",
    "verylongfermata",
]
"""A sequence of all allowed fermata types.

'allowed' means merely that all defined strings here
should work with the `abjad` module in mutwos frontend module."""

FERMATA_TYPE_LITERAL = typing.Literal[
    "shortfermata",
    "fermata",
    "longfermata",
    "verylongfermata",
]
"""A sequence of all allowed fermata types.

'allowed' means merely that all defined strings here
should work with the `abjad` module in mutwos frontend module."""

HAIRPIN_SYMBOL_LITERAL = typing.Literal["<", ">", "!"]
"""A sequence of all allowed hairpin symbols.

'allowed' means merely that all defined strings here
should work with the `abjad` module in mutwos frontend module."""

DIRECTION_LITERAL = typing.Literal["up", "down"]
"""A sequence of all allowed directions.

'allowed' means merely that all defined strings here
should work with the `abjad` module in mutwos frontend module."""



"""Generic decorators that are used within :mod:`mutwo`."""

import copy
import functools
import os
import typing

try:
    import dill as pickle
except ImportError:
    import pickle

__all__ = ("add_copy_option", "compute_lazy")


F = typing.TypeVar("F", bound=typing.Callable[..., typing.Any])


def add_copy_option(function: F) -> F:
    """This decorator adds a return option for object mutating methods.

    :arg function: The method which shall be adjusted.

    The 'add_copy_option' decorator adds the 'mutate' keyword argument
    to the decorated method. If 'mutate' is set to False, the decorator deep
    copies the respective object, then applies the called method on the new
    copied object and finally returns the copied object. This can be useful
    for methods that by default mutate its object. When adding this method,
    it is up to the user whether the original object shall be changed
    and returned (for mutate=True) or if a copied version of the object with
    the respective mutation shall be returned (for mutate=False).
    """

    @functools.wraps(function)
    def wrapper(self, *args, mutate: bool = True, **kwargs) -> typing.Any:
        if mutate is True:
            function(self, *args, **kwargs)
            return self
        else:
            deep_copied_object = copy.deepcopy(self)
            function(deep_copied_object, *args, **kwargs)
            return deep_copied_object

    wrapped_function = typing.cast(F, wrapper)
    wrapped_function.__annotations__.update({"mutate": bool})
    return wrapped_function


G = typing.TypeVar("G")


def add_tag_to_class(class_to_decorate: G) -> G:
    """This decorator adds a 'tag' argument to the init method of a class.

    :arg class_to_decorate: The class which shall be decorated.
    """

    init = class_to_decorate.__init__

    def init_with_tag(self, *args, tag: typing.Optional[str] = None, **kwargs):
        init(self, *args, **kwargs)
        self.tag = tag

    class_to_decorate.__init__ = init_with_tag
    return class_to_decorate


def compute_lazy(path: str, force_to_compute: bool = False):
    """Only run function if its input changes and otherwise load return value from disk.

    :param path: Where to save the computed result.
    :type path: str
    :param force_to_compute: Set to ``True`` if function has to be re-computed.
    :type force_to_compute: bool

    This function is helpful if there is a complex, long-taking calculation,
    which should only run once or from time to time if the input changes.

    **Example:**

    >>> from mutwo.utilities import decorators
    >>> @decorators.compute_lazy("magic_output", False)
        def my_super_complex_calculation(n_numbers):
            return sum(number for number in range(n_numbers))
    >>> N_NUMBERS = 100000000
    >>> my_super_complex_calculation(N_NUMBERS)
    4999999950000000
    >>> # takes very little time when calling the function the second time
    >>> my_super_complex_calculation(N_NUMBERS)
    4999999950000000
    >>> # takes long again, because the input changed
    >>> my_super_complex_calculation(N_NUMBERS + 10)
    4999999950000000
    """

    def decorator(function_to_decorate: F) -> F:
        @functools.wraps(function_to_decorate)
        def wrapper(*args, **kwargs) -> typing.Any:
            has_to_compute = False

            current_args_and_kwargs = (args, kwargs)
            is_file = os.path.isfile(path)

            if not is_file:
                has_to_compute = True
            else:
                with open(path, "rb") as f:
                    function_result, previous_args_and_kwargs = pickle.load(f)

                if previous_args_and_kwargs != current_args_and_kwargs:
                    has_to_compute = True

            if has_to_compute or force_to_compute:
                function_result = function_to_decorate(*args, **kwargs)
                with open(path, "wb") as f:
                    pickle.dump((function_result, current_args_and_kwargs), f)

            return function_result

        wrapped_function = typing.cast(F, wrapper)
        return wrapped_function

    return decorator



"""Submodule for the parameter tempo.
"""

import typing

from mutwo.core.utilities import constants

__all__ = ("TempoPoint",)

TempoInBeatsPerMinute = float
TempoRangeInBeatsPerMinute = tuple[TempoInBeatsPerMinute, TempoInBeatsPerMinute]
TempoOrTempoRangeInBeatsPerMinute = typing.Union[
    TempoInBeatsPerMinute, TempoRangeInBeatsPerMinute
]


class TempoPoint(object):
    def __init__(
        self,
        tempo_or_tempo_range_in_beats_per_minute: TempoOrTempoRangeInBeatsPerMinute,
        reference: constants.Real = 1,
        textual_indication: typing.Optional[str] = None,
    ):
        self.tempo_or_tempo_range_in_beats_per_minute = (
            tempo_or_tempo_range_in_beats_per_minute
        )
        self.reference = reference
        self.textual_indication = textual_indication

    def __repr__(self) -> str:
        return "{}(BPM: {}, reference: {})".format(
            type(self).__name__, self.tempo_in_beats_per_minute, self.reference
        )

    def __eq__(self, other: object) -> bool:
        attributes_to_compare = (
            "tempo_in_beats_per_minute",
            "reference",
            "textual_indication",
        )
        try:
            comparision_tuple = tuple(
                getattr(other, attribute) for attribute in attributes_to_compare
            )
        except AttributeError:
            return False

        return all(
            tuple(
                getattr(self, attribute) == value_of_other
                for attribute, value_of_other in zip(
                    attributes_to_compare, comparision_tuple
                )
            )
        )

    @property
    def tempo_in_beats_per_minute(self) -> TempoInBeatsPerMinute:
        if isinstance(self.tempo_or_tempo_range_in_beats_per_minute, tuple):
            return self.tempo_or_tempo_range_in_beats_per_minute[0]
        else:
            return self.tempo_or_tempo_range_in_beats_per_minute

    @property
    def absolute_tempo_in_beat_per_minute(self) -> float:
        return self.tempo_in_beats_per_minute * self.reference



"""Several constants which are used for the loudness converter module.

"""

AUDITORY_THRESHOLD_AT_1KHZ = 0.00002
"""Roughly the sound of a mosquito flying 3 m away
(see https://en.wikipedia.org/wiki/Sound_pressure)."""



"""Apply grace notes on Events"""

import copy
import typing

import expenvelope

from mutwo.core import converters
from mutwo.core import events
from mutwo.core import utilities

__all__ = ("GraceNotesConverter",)


class GraceNotesConverter(converters.abc.EventConverter):
    """Apply grace notes and after grace notes on :class:`events.abc.Event`.

    :param minima_grace_notes_duration_factor: Minimal percentage how much
        of the initial duration of the :class:`~mutwo.events.basic.SimpleEvent`
        shall be moved to the grace notes / after grace notes. This value has to be
        smaller than 0.5 (so that the :class:`SimpleEvent` have a
        duration > 0 if it has both: grace notes and after grace notes)
        and bigger than 0 (so that the grace notes or after grace notes
        have a duration > 0). Default to 0.12.
    :type minima_grace_notes_duration_factor: float
    :param maxima_grace_notes_duration_factor: Maxima percentage how much
        of the initial duration of the :class:`~mutwo.events.basic.SimpleEvent`
        shall be moved to the grace notes / after grace notes. This value has to be
        smaller than 0.5 (so that the :class:`SimpleEvent` have a
        duration > 0 if it has both: grace notes and after grace notes)
        and bigger than 0 (so that the grace notes or after grace notes
        have a duration > 0). Default to 0.25.
    :type maxima_grace_notes_duration_factor: float
    :param minima_number_of_grace_notes: For how many events in the grace
        note or after grace note container shall the
        `minima_grace_notes_duration_factor` be applied. Default to 1.
    :type minima_number_of_grace_notes: int
    :param maxima_number_of_grace_notes: For how many events in the grace
        note or after grace note container shall the
        `maxima_number_of_grace_notes` be applied. Default to 4.
    :type maxima_number_of_grace_notes: int
    :param simple_event_to_grace_note_sequential_event: Function which
        receives as an input a :class:`~mutwo.events.basic.SimpleEvent`
        and returns a :class:`~mutwo.events.basic.SequentialEvent`.
        By default the function will ask the event for a
        :attr:`~mutwo.events.music.NoteLike.grace_note_sequential_event`
        attribute, because by default `~mutwo.events.music.NoteLike`
        objects are expected.
    :type simple_event_to_grace_note_sequential_event: typing.Callable[[events.basic.SimpleEvent], events.basic.SequentialEvent[events.basic.SimpleEvent]]
    :param simple_event_to_after_grace_note_sequential_event: Function which
        receives as an input a :class:`~mutwo.events.basic.SimpleEvent`
        and returns a :class:`~mutwo.events.basic.SequentialEvent`.
        By default the function will ask the event for a
        :attr:`~mutwo.events.music.NoteLike.grace_note_sequential_event`
        attribute, because by default `~mutwo.events.music.NoteLike`
        objects are expected.
    :type simple_event_to_after_grace_note_sequential_event: typing.Callable[[events.basic.SimpleEvent], events.basic.SequentialEvent[events.basic.SimpleEvent]]
    """

    def __init__(
        self,
        minima_grace_notes_duration_factor: float = 0.12,
        maxima_grace_notes_duration_factor: float = 0.25,
        minima_number_of_grace_notes: int = 1,
        maxima_number_of_grace_notes: int = 4,
        simple_event_to_grace_note_sequential_event: typing.Callable[
            [events.basic.SimpleEvent],
            events.basic.SequentialEvent[events.basic.SimpleEvent],
        ] = lambda simple_event: simple_event.grace_note_sequential_event,
        simple_event_to_after_grace_note_sequential_event: typing.Callable[
            [events.basic.SimpleEvent],
            events.basic.SequentialEvent[events.basic.SimpleEvent],
        ] = lambda simple_event: simple_event.after_grace_note_sequential_event,
    ):
        self._test_input(
            minima_grace_notes_duration_factor,
            maxima_grace_notes_duration_factor,
            minima_number_of_grace_notes,
            maxima_number_of_grace_notes,
        )

        self._simple_event_to_grace_note_sequential_event = (
            simple_event_to_grace_note_sequential_event
        )
        self._simple_event_to_after_grace_note_sequential_event = (
            simple_event_to_after_grace_note_sequential_event
        )

        self._n_grace_notes_to_grace_note_duration_factor_envelope = (
            expenvelope.Envelope.from_points(
                (minima_number_of_grace_notes, minima_grace_notes_duration_factor),
                (maxima_number_of_grace_notes, maxima_grace_notes_duration_factor),
            )
        )

    @staticmethod
    def _test_input(
        minima_grace_notes_duration_factor: float,
        maxima_grace_notes_duration_factor: float,
        minima_number_of_grace_notes: int,
        maxima_number_of_grace_notes: int,
    ):
        try:
            assert minima_number_of_grace_notes < maxima_number_of_grace_notes
        except AssertionError:
            message = "'minima_number_of_grace_notes' has to be smaller than 'maxima_number_of_grace_notes'!"
            raise ValueError(message)

        try:
            assert (
                minima_grace_notes_duration_factor < maxima_grace_notes_duration_factor
            )
        except AssertionError:
            message = "'minima_grace_notes_duration_factor' has to be smaller than 'maxima_grace_notes_duration_factor'!"
            raise ValueError(message)

        try:
            assert maxima_grace_notes_duration_factor < 0.5
        except AssertionError:
            message = "'maxima_grace_notes_duration_factor' has to be smaller than 0.5!"
            raise ValueError(message)

        try:
            assert minima_grace_notes_duration_factor > 0
        except AssertionError:
            message = "'minima_grace_notes_duration_factor' has to be bigger than 0!"
            raise ValueError(message)

    @staticmethod
    def _get_before_or_after_grace_note_sequential_event(
        simple_event_to_convert: events.basic.SimpleEvent,
        simple_event_to_before_or_after_grace_note_sequential_event: typing.Callable[
            [events.basic.SimpleEvent],
            events.basic.SequentialEvent[events.basic.SimpleEvent],
        ],
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        return utilities.tools.call_function_except_attribute_error(
            simple_event_to_before_or_after_grace_note_sequential_event,
            simple_event_to_convert,
            events.basic.SequentialEvent([]),
        )

    def _get_grace_note_sequential_event(
        self, simple_event_to_convert: events.basic.SimpleEvent
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        return GraceNotesConverter._get_before_or_after_grace_note_sequential_event(
            simple_event_to_convert, self._simple_event_to_grace_note_sequential_event
        )

    def _get_after_grace_note_sequential_event(
        self, simple_event_to_convert: events.basic.SimpleEvent
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        return GraceNotesConverter._get_before_or_after_grace_note_sequential_event(
            simple_event_to_convert,
            self._simple_event_to_after_grace_note_sequential_event,
        )

    def _convert_simple_event(
        self,
        event_to_convert: events.basic.SimpleEvent,
        absolute_entry_delay: utilities.constants.DurationType,
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        """Convert instance of :class:`mutwo.events.basic.SimpleEvent`."""

        def adjust_grace_note_sequential_event(
            grace_note_sequential_event: events.basic.SequentialEvent[
                events.basic.SimpleEvent
            ],
        ) -> events.basic.SequentialEvent:
            if grace_note_sequential_event:
                factor_to_main_event = (
                    self._n_grace_notes_to_grace_note_duration_factor_envelope.value_at(
                        len(grace_note_sequential_event)
                    )
                )
                new_duration = event_to_convert.duration * factor_to_main_event
                grace_note_sequential_event = grace_note_sequential_event.copy()
                grace_note_sequential_event.duration = new_duration
            return grace_note_sequential_event

        grace_note_sequential_event = adjust_grace_note_sequential_event(
            self._get_grace_note_sequential_event(event_to_convert)
        )
        after_grace_note_sequential_event = adjust_grace_note_sequential_event(
            self._get_after_grace_note_sequential_event(event_to_convert)
        )

        copied_event_to_convert = copy.deepcopy(event_to_convert)
        # Remove applied grace notes / after grace notes
        copied_event_to_convert.grace_note_sequential_event = (
            events.basic.SequentialEvent([])
        )
        copied_event_to_convert.after_grace_note_sequential_event = (
            events.basic.SequentialEvent([])
        )
        copied_event_to_convert.duration -= (
            grace_note_sequential_event.duration
            + after_grace_note_sequential_event.duration
        )

        grace_note_sequential_event.append(copied_event_to_convert)
        grace_note_sequential_event.extend(after_grace_note_sequential_event)

        return grace_note_sequential_event

    def _convert_simultaneous_event(
        self,
        simultaneous_event: events.basic.SimultaneousEvent,
        absolute_entry_delay: utilities.constants.DurationType,
    ) -> tuple[events.basic.SimultaneousEvent[events.abc.Event]]:
        """Convert instance of :class:`mutwo.events.basic.SimultaneousEvent`."""

        simultaneous_event_copied = simultaneous_event.empty_copy()
        for event in simultaneous_event:
            converted_event = self._convert_event(event, absolute_entry_delay)
            # If we find a simple event, we shouldn't extend but append it to the
            # the SimultaneousEvent. A converted simple event will be a
            # SequentialEvent. The grace notes and after grace notes should
            # happen before and after the respective event. If we extend
            # the whole SequentialEvent to the SimultaneousEvent the
            # grace notes and after grace notes will be played simultaneously
            # with the main note, therefore we have to append them.
            if isinstance(event, events.basic.SimpleEvent):
                simultaneous_event_copied.append(converted_event)
            else:
                simultaneous_event_copied.extend(converted_event)
        return (simultaneous_event_copied,)

    def _convert_sequential_event(
        self,
        sequential_event: events.basic.SequentialEvent,
        absolute_entry_delay: utilities.constants.DurationType,
    ) -> tuple[events.basic.SequentialEvent[events.abc.Event]]:
        sequential_event_copied = sequential_event.empty_copy()
        sequential_event_copied.extend(
            super()._convert_sequential_event(sequential_event, absolute_entry_delay)
        )
        return (sequential_event_copied,)

    def convert(self, event_to_convert: events.abc.Event) -> events.abc.Event:
        """Apply grace notes and after grace notes of all :class:`SimpleEvent`.

        :param event_to_convert: The event which grace notes and after grace
            notes shall be converted to normal events in the upper
            :class:`SequentialEvent`.
        :type event_to_convert: events.abc.Event
        """

        converted_event = self._convert_event(event_to_convert, 0)
        if isinstance(event_to_convert, events.basic.SimpleEvent):
            return converted_event
        else:
            return converted_event[0]



"""Module for mutwo specific exceptions."""

import typing

from mutwo.core.utilities import constants

__all__ = (
    "ValueNotAssignedError",
    "OverlappingTimeBracketsError",
    "SplitUnavailableChildError",
    "NoSolutionFoundError",
)


class ValueNotAssignedError(Exception):
    def __init__(self):
        super().__init__(
            "TimeBracket values haven't been assigned yet, run"
            " 'assign_concrete_times' method first"
        )


class OverlappingTimeBracketsError(Exception):
    def __init__(
        self,
        previous_end_time: typing.Optional[float] = None,
        start_time: typing.Optional[float] = None,
        time_bracket=None,
    ):
        message = "Found overlapping time brackets!"
        if previous_end_time and start_time:
            previous_end_time = round(previous_end_time / 60, 2)
            start_time = round(previous_end_time / 60, 2)
            message += f"\nThe end time of the previous time bracket is {previous_end_time}. The start time of the current time bracket is {start_time}."
        if time_bracket:
            message += f"\nThe problematic candidate is:\n{repr(time_bracket)[:50]}."
        super().__init__(message)


class SplitUnavailableChildError(Exception):
    def __init__(self, absolute_time: constants.DurationType):
        super().__init__(
            f"Can't split child at absolute time '{absolute_time}'. There is no child"
            " event available at the requested time."
        )


class NoSolutionFoundError(Exception):
    def __init__(self, message: str):
        super().__init__(message)



"""Envelope events"""

from __future__ import annotations

import typing

from scipy import integrate

from mutwo.core import events
from mutwo.core.utilities import constants
from mutwo.core.utilities import tools


__all__ = ("Envelope", "RelativeEnvelope")

T = typing.TypeVar("T", bound=events.abc.Event)


class Envelope(events.basic.SequentialEvent, typing.Generic[T]):
    """Model continuous changing values (e.g. glissandi, crescendo).

    :param event_iterable_or_point_sequence: An iterable filled with events. Each event represents
        a point in a two dimensional graph where the y-axis presents time
        and the x-axis a changing value. Any event class can be used. It is
        more important that the used event classes fit with the functions
        passed in the following parameters.
    :type event_iterable_or_point_sequence: typing.Iterable[T]
    :param event_to_parameter: A function which receives an event and has to
        return a parameter object (any object). By default the function will
        ask the event for its `value` property. If the property can't be found
        it will return 0.
    :type event_to_parameter: typing.Callable[[events.abc.Event], constants.ParameterType]
    :param event_to_curve_shape: A function which receives an event and has
        to return a curve_shape. A curve_shape is either a float, an integer
        or a fraction. For a curve_shape = 0 a linear transition between two
        points is created. For a curve_shape > 0 the envelope changes slower
        at the beginning and faster at the end, for a curve_shape < 0 it is
        the inverse behaviour. The default function will ask the event for its
        `curve_shape` property. If the property can't be found
        it will return 0.
    :type event_to_curve_shape: typing.Callable[[events.abc.Event], CurveShape]
    :param parameter_to_value: Convert a parameter to a value. A value is any
        object which supports mathematical operations.
    :type parameter_to_value: typing.Callable[[Value], constants.ParameterType]
    :param value_to_parameter:
    :type value_to_parameter: typing.Callable[[Value], constants.ParameterType]
    :param apply_parameter_on_event:
    :type apply_parameter_on_event: typing.Callable[[events.abc.Event, constants.ParameterType], None]
    :param apply_curve_shape_on_event:
    :type apply_curve_shape_on_event: typing.Callable[[events.abc.Event, CurveShape], None]
    :param default_event_class:
    :type default_event_class: type[events.abc.Event]
    :param initialise_default_event_class:
    :type initialise_default_event_class: typing.Callable[[type[events.abc.Event], constants.DurationType], events.abc.Event]

    This class is inspired by Marc Evanstein `Envelope` class in his
    `expenvelope <https://git.sr.ht/~marcevanstein/expenvelope>`_
    python package and is made to fit better into the `mutwo` ecosystem.
    """

    # Type definitions
    Value = constants.Real
    CurveShape = constants.Real
    IncompletePoint = tuple[constants.DurationType, constants.ParameterType]
    CompletePoint = tuple[
        constants.DurationType, constants.ParameterType, CurveShape  # type: ignore
    ]
    Point = typing.Union[CompletePoint, IncompletePoint]

    # class specific side attributes (keep consistent
    # when applying the "empty_copy" method)
    _class_specific_side_attribute_tuple = (
        "event_to_parameter",
        "event_to_curve_shape",
        "value_to_parameter",
        "parameter_to_value",
        "apply_parameter_on_event",
        "apply_curve_shape_on_event",
        "default_event_class",
        "initialise_default_event_class",
    )

    def __init__(
        self,
        event_iterable_or_point_sequence: typing.Union[
            typing.Iterable[T], typing.Sequence[Point]
        ],
        event_to_parameter: typing.Callable[
            [events.abc.Event], constants.ParameterType
        ] = lambda event: getattr(
            event, events.envelopes_constants.DEFAULT_PARAMETER_ATTRIBUTE_NAME
        )
        if hasattr(event, events.envelopes_constants.DEFAULT_PARAMETER_ATTRIBUTE_NAME)
        else 0,
        event_to_curve_shape: typing.Callable[
            [events.abc.Event], CurveShape
        ] = lambda event: getattr(
            event, events.envelopes_constants.DEFAULT_CURVE_SHAPE_ATTRIBUTE_NAME
        )
        if hasattr(event, events.envelopes_constants.DEFAULT_CURVE_SHAPE_ATTRIBUTE_NAME)
        else 0,
        parameter_to_value: typing.Callable[
            [Value], constants.ParameterType
        ] = lambda parameter: parameter,
        value_to_parameter: typing.Callable[
            [Value], constants.ParameterType
        ] = lambda value: value,
        apply_parameter_on_event: typing.Callable[
            [events.abc.Event, constants.ParameterType], None
        ] = lambda event, parameter: setattr(
            event,
            events.envelopes_constants.DEFAULT_PARAMETER_ATTRIBUTE_NAME,
            parameter,
        ),
        apply_curve_shape_on_event: typing.Callable[
            [events.abc.Event, CurveShape], None
        ] = lambda event, curve_shape: setattr(
            event,
            events.envelopes_constants.DEFAULT_CURVE_SHAPE_ATTRIBUTE_NAME,
            curve_shape,
        ),
        default_event_class: type[events.abc.Event] = events.basic.SimpleEvent,
        initialise_default_event_class: typing.Callable[
            [type[events.abc.Event], constants.DurationType], events.abc.Event
        ] = lambda simple_event, duration: simple_event(duration),
    ):
        self.event_to_parameter = event_to_parameter
        self.event_to_curve_shape = event_to_curve_shape
        self.value_to_parameter = value_to_parameter
        self.parameter_to_value = parameter_to_value
        self.apply_parameter_on_event = apply_parameter_on_event
        self.apply_curve_shape_on_event = apply_curve_shape_on_event
        self.default_event_class = default_event_class
        self.initialise_default_event_class = initialise_default_event_class

        item_type_list = [
            isinstance(event_or_point, events.abc.Event)
            for event_or_point in event_iterable_or_point_sequence
        ]
        if all(item_type_list):
            event_iterable = event_iterable_or_point_sequence
        elif any(item_type_list):
            raise TypeError(
                "Found inconsistent iterable with mixed types. "
                "Please only use events or only use points for "
                "'event_iterable_or_point_sequence'. First 200 "
                "characters of the problematic iterable: \n"
                f"{str(event_iterable_or_point_sequence)[:200]}"
            )
        else:
            event_iterable = self._point_sequence_to_event_list(
                event_iterable_or_point_sequence  # type: ignore
            )
        super().__init__(event_iterable)

    # ###################################################################### #
    #                      public class methods                              #
    # ###################################################################### #

    @classmethod
    def from_points(
        cls,
        *point: Point,
        **kwargs,
    ) -> Envelope:
        return cls(point, **kwargs)

    # ###################################################################### #
    #                    private static methods                              #
    # ###################################################################### #

    @staticmethod
    def _point_sequence_to_corrected_point_list(
        point_or_invalid_type_sequence: typing.Sequence[typing.Union[Point, typing.Any]]
    ) -> list[typing.Union[Envelope.CompletePoint, None]]:
        corrected_point_list: list[typing.Union[Envelope.CompletePoint, None]] = []
        for point in point_or_invalid_type_sequence:
            point_length = len(point)
            if point_length == 2:
                point += (0,)  # type: ignore
            elif point_length != 3:
                raise TypeError(
                    f"Found invalid point: '{point}' with {point_length} "
                    "items! Points should have two or three items."
                )
            corrected_point_list.append(point)  # type: ignore
        return corrected_point_list

    # ###################################################################### #
    #                         private methods                                #
    # ###################################################################### #

    def _point_sequence_to_event_list(
        self,
        point_or_invalid_type_sequence: typing.Sequence[
            typing.Union[Point, typing.Any]
        ],
    ) -> list[events.abc.Event]:
        corrected_point_list = Envelope._point_sequence_to_corrected_point_list(
            point_or_invalid_type_sequence
        )
        corrected_point_list.append(None)
        event_list = []
        for point0, point1 in zip(corrected_point_list, corrected_point_list[1:]):
            if point0 is not None:
                absolute_time0, value_or_parameter, curve_shape = point0
            else:
                raise TypeError("Found unexpected position of None in provided points.")
            if point1:
                absolute_time1 = point1[0]
                assert absolute_time1 >= absolute_time0
            else:
                absolute_time1 = absolute_time0
            duration = absolute_time1 - absolute_time0
            event = self.initialise_default_event_class(
                self.default_event_class, duration
            )
            self.apply_parameter_on_event(event, value_or_parameter)
            self.apply_curve_shape_on_event(event, curve_shape)
            event_list.append(event)
        return event_list

    def _event_to_value(self, event: events.abc.Event) -> Value:
        return self.parameter_to_value(self.event_to_parameter(event))

    # ###################################################################### #
    #                         public properties                              #
    # ###################################################################### #

    @property
    def value_tuple(self) -> tuple[Value, ...]:
        return tuple(map(self._event_to_value, self))

    @property
    def is_static(self) -> bool:
        """Return `True` if :class:`Envelope` only has one static value."""

        return len(set(self.value_tuple)) <= 1

    # ###################################################################### #
    #                          public methods                                #
    # ###################################################################### #

    def value_at(self, absolute_time: constants.DurationType) -> Value:
        absolute_time_tuple = self.absolute_time_tuple

        use_only_first_event = absolute_time <= absolute_time_tuple[0]
        use_only_last_event = absolute_time >= absolute_time_tuple[-1]
        if use_only_first_event or use_only_last_event:
            index = 0 if use_only_first_event else -1
            return self._event_to_value(self[index])

        event_0_index = self._get_index_at_from_absolute_time_tuple(
            absolute_time, absolute_time_tuple, self.duration
        )
        assert event_0_index is not None

        value0, value1 = (
            self._event_to_value(self[event_0_index + n]) for n in range(2)
        )
        curve_shape = self.event_to_curve_shape(self[event_0_index])

        return tools.scale(
            absolute_time,
            absolute_time_tuple[event_0_index],
            absolute_time_tuple[event_0_index + 1],
            value0,
            value1,
            curve_shape,
        )

    def parameter_at(
        self, absolute_time: constants.DurationType
    ) -> constants.ParameterType:
        return self.value_to_parameter(self.value_at(absolute_time))

    def integrate_interval(
        self, start: constants.DurationType, end: constants.DurationType
    ) -> float:
        return integrate.quad(lambda x: self.value_at(x), start, end)[0]

    def get_average_value(
        self,
        start: typing.Optional[constants.DurationType] = None,
        end: typing.Optional[constants.DurationType] = None,
    ) -> Value:
        if start is None:
            start = 0
        if end is None:
            end = self.duration
        duration = end - start
        if duration == 0:
            return 0
        return self.integrate_interval(start, end) / duration

    def get_average_parameter(
        self,
        start: typing.Optional[constants.DurationType] = None,
        end: typing.Optional[constants.DurationType] = None,
    ) -> constants.ParameterType:
        return self.value_to_parameter(self.get_average_value(start, end))


class RelativeEnvelope(Envelope, typing.Generic[T]):
    __parent_doc_string = Envelope.__doc__.split("\n")[2:]  # type: ignore
    __after_parameter_text_index = __parent_doc_string.index("")
    __doc__ = "\n".join(
        ["Envelope with relative durations and values / parameters.\n"]
        + __parent_doc_string[:__after_parameter_text_index]
        + [
            "    :param base_parameter_and_relative_parameter_to_absolute_parameter: A function",
            "        which runs when the :func:`resolve` is called. It expects the base parameter",
            "        and the relative parameter (which is extracted from the envelope events)",
            "        and should return an absolute parameter.",
        ]
        + __parent_doc_string[__after_parameter_text_index:]
        + [
            "    The :class:`RelativeEnvelope` adds the :func:`resolve` method",
            "    to the base class :class:`Envelope`.",
        ]
    )

    def __init__(
        self,
        *args,
        base_parameter_and_relative_parameter_to_absolute_parameter: typing.Callable[
            [constants.ParameterType, constants.ParameterType], constants.ParameterType
        ],
        **kwargs,
    ):
        self.base_parameter_and_relative_parameter_to_absolute_parameter = (
            base_parameter_and_relative_parameter_to_absolute_parameter
        )
        super().__init__(*args, **kwargs)

    def resolve(
        self,
        duration: constants.DurationType,
        base_parameter: constants.ParameterType,
        resolve_envelope_class: type[Envelope] = Envelope,
    ) -> Envelope:
        point_list = []
        try:
            duration_factor = duration / self.duration
        except ZeroDivisionError:
            duration_factor = 0
        for absolute_time, event in zip(self.absolute_time_tuple, self):
            relative_parameter = self.event_to_parameter(event)
            new_parameter = (
                self.base_parameter_and_relative_parameter_to_absolute_parameter(
                    base_parameter, relative_parameter
                )
            )
            point = (
                absolute_time * duration_factor,
                new_parameter,
                self.event_to_curve_shape(event),
            )
            point_list.append(point)
        return resolve_envelope_class(point_list)



"""Abstractions of `tuning commas <https://en.wikipedia.org/wiki/Comma_(music)>`_

The tuning commas are helpful for :class:`mutwo.parameters.pitches.JustIntonationPitch`
which owns a :attr:`commas` attribute. By default :mod:`mutwo` makes use of commas
defined by the
`Helmholtz-Ellis JI Pitch Notation <https://marsbat.space/pdfs/notation.pdf>`_.
"""

import functools
import operator
import typing

try:
    import quicktions as fractions  # type: ignore
except ImportError:
    import fractions  # type: ignore


class Comma(object):
    """A `tuning comma <https://en.wikipedia.org/wiki/Comma_(music)>`_."""

    def __init__(self, ratio: fractions.Fraction):
        self._ratio = ratio

    def __repr__(self) -> str:
        return "Comma({})".format(self.ratio)

    @property
    def ratio(self) -> fractions.Fraction:
        return self._ratio


class CommaCompound(typing.Iterable[Comma]):
    """Collection of tuning commas."""

    # basically a frozen dict?

    def __init__(
        self,
        prime_to_exponent_dict: dict[int, int],
        prime_to_comma_dict: typing.Optional[dict[int, Comma]],
    ):
        # TODO(make sure all primes in 'prime_to_exponent_dict' are also in
        # 'prime_to_comma_dict')

        self._prime_to_exponent_dict = prime_to_exponent_dict
        self._prime_to_comma_dict = prime_to_comma_dict

    def __repr__(self) -> str:
        return "{}({})".format(type(self).__name__, self._prime_to_exponent_dict)

    def __len__(self) -> int:
        return sum((abs(exponent) for exponent in self._prime_to_exponent_dict.values()))

    def __iter__(self):
        return (
            self._prime_to_comma_dict[prime].ratio ** exponent
            for prime, exponent in self._prime_to_exponent_dict.items()
        )

    @property
    def prime_to_exponent_dict(self) -> dict[int, int]:
        return dict(self._prime_to_exponent_dict)

    @property
    def ratio(self) -> fractions.Fraction:
        if self.prime_to_exponent_dict:
            return functools.reduce(operator.mul, iter(self))
        else:
            return fractions.Fraction(1, 1)



from mutwo.core.utilities.tools import fetch_doc_string_from_core_module
fetch_doc_string_from_core_module()
del fetch_doc_string_from_core_module

from . import abc
from . import symmetrical
from . import shared
from . import frontends
from . import backends
